#ifndef BIT_ARRAY_HH
#define BIT_ARRAY_HH

#include <errno.h>
#include "bit/Buffer.hh"

namespace bit {

  /** A bit-packed array of fixed-bit-width elements. 
   *
   * \bug Growing the array may have some problems when the number of
   * elements is larger than 2^63.  For example, when the capacity is
   * doubled in set_grow().
   */
  class Array {

  public:
    /** A limit index value used for return values.  All real indices
     * must be less than this value. */
    static const u64 index_limit = (u64)-1;

  public:
    
    /** Construct an empty array with zero bit-width. */
    Array()
      : m_bits_per_elem(0), m_num_elems(0), m_capacity(0)
    {
    }

    /** Create an array with initial size.  The initial elements are
     * guaranteed to be zero.
     *
     * \param bits_per_elem = bits_per_element (0-32 bits)
     * \param num_elems = initial number of elements in the array
     * \throw bit::invalid_argument one of the bit arguments is invalid
     */
    Array(u64 num_elems, unsigned int bits_per_elem)
      : m_num_elems(0), m_capacity(0)
    {
      if (bits_per_elem > max_bits_per_value)
        throw bit::invalid_argument("bit::Array(): too many bits per elem");
      m_bits_per_elem = bits_per_elem;
      resize(num_elems);
    }

    /** Return the number of elements in the array. */
    u64 num_elems() const { 
      return m_num_elems;
    }

    /** Return the width of the array, i.e. the number of bits
     * allocated for each element. */
    unsigned int bits_per_elem() const {
      return m_bits_per_elem;
    }

    /** Return the number of elements allocated for the underlying buffer. */
    u64 capacity() const {
      return m_capacity;
    }

    /** Change the number of elements in the array.  The values of the
     * possible new elements are guaranteed to be zero ONLY if the
     * array has never been resized smaller.  
     * \param num_elems = the new number of elements
     */
    void resize(u64 num_elems)
    {
      m_num_elems = num_elems;
      if (m_capacity < m_num_elems) {
        m_capacity = m_num_elems;
        m_buffer.resize(bytes_required(m_capacity, m_bits_per_elem));
      }
    }

    /** Change the capacity (number of allocated elements) of the
     * array.  Making the capacity smaller does nothing.
     * \param capacity = the number of elements to allocate
     */
    void reserve(u64 capacity)
    {
      if (capacity <= m_capacity)
        return;
      m_capacity = capacity;
      m_buffer.resize(bytes_required(m_capacity, m_bits_per_elem));
    }

    /** Change the width of the elements to given width.  All elements
     * must fit in the given width or bit::invalid_argument will be thrown and
     * the array is guaranteed to be left in the original state.
     * \param bits_per_elem = the number of bits to use per element (1-32 bits)
     * \throw bit::invalid_argument invalid bits_per_elem
     */
    void set_width(unsigned int bits_per_elem)
    {
      if (bits_per_elem == m_bits_per_elem)
        return;

      Buffer new_buffer(bytes_required(m_capacity, bits_per_elem));
      for (u64 i = 0; i < m_num_elems; i++) {
        u32 value = get(i);
        new_buffer.write_elem(i, bits_per_elem, value);
      }
      m_buffer.swap(new_buffer);
      m_bits_per_elem = bits_per_elem;
    }

    /** Set the value of an element.  
     *
     * \param elem = the index of the element to set
     * \param value = the value to set
     * \throw bit::invalid_argument value wider than the bit-width of
     * the array or trying to write to zero-width array
     * \throw bit::out_of_range accessing outside the array
     */
    void set(u64 elem, u32 value)
    {
      if (elem >= m_num_elems)
        throw bit::out_of_range("bit::Array::set(): out of range");
      m_buffer.write_elem(elem, m_bits_per_elem, value);
    }

    /** Set the value of an element growing the buffer if necessary.
     * If the current capacity is not enough, the capacity is doubled
     * (set to one from zero), and if that is not enough, the capacity
     * is set to (elem+1).
     *
     * \warning Reserving capacity and resizing the array happens
     * before the actual value is written to the buffer.  Thus, the
     * size and capacity of the array may change even if the write
     * throws bit::invalid_argument for too wide write.
     *
     * \param elem = the index of the element to set
     * \param value = the value to set
     * \throw bit::invalid_argument value wider than the bit-width of the array
     */
    void set_grow(u64 elem, u32 value)
    {
      if (elem >= m_num_elems) {
        if (elem >= m_capacity) {
          u64 new_capacity = m_capacity * 2;
          if (elem >= new_capacity)
            new_capacity = elem + 1;
          reserve(new_capacity);
        }
        resize(elem + 1);
      }
      m_buffer.write_elem(elem, m_bits_per_elem, value);
    }

    /** Set the value of an element growing and widening the buffer if
     * necessary.  See set_grow() for info how capacity is handled and
     * warnings about exceptions.
     *
     * \param elem = the index of the element to set
     * \param value = the value to set
     */
    void set_grow_widen(u64 elem, u32 value)
    {
      unsigned int bits = highest_bit(value);
      if (bits > m_bits_per_elem)
        set_width(bits);
      set_grow(elem, value);
    }

    /** Return value of an element. 
     * \param elem = the index of the element
     * \return the value
     * \throw bit::out_of_range accessing outside the array
     */
    u32 get(u64 elem) const
    {
      if (elem >= m_num_elems)
        throw bit::out_of_range("bit::Array::get(): out of range");
      return m_buffer.read_elem(elem, m_bits_per_elem);
    }

    /** Const access the internal data buffer.  It is safe to write
     * and read the first data_len() bytes of the pointer.  Note that
     * set_grow(), set_grow_widen(), set_width(), resize() and
     * reserve() may invalidate the pointer. */
    const unsigned char *data() const
    {
      return m_buffer.data();
    }

    /** Access the internal data buffer.  It is safe to write
     * and read the first data_len() bytes of the pointer.  Note that
     * set_grow(), set_grow_widen(), set_width(), resize() and
     * reserve() may invalidate the pointer. */
    unsigned char *data() 
    {
      return m_buffer.data();
    }

    /** Return the safe access length of the internal data buffer
     * obtained with data() (i.e. minimum number of bytes needed to
     * store the elements of the array. */
    u64 data_len() const 
    {
      return bytes_required(m_num_elems, m_bits_per_elem);
    }

    /** Write the array in file.  Note that capacity is not written
     * into the file.  When the array is read, the capacity is set to
     * number of elements.
     *
     * \param file = file stream to write to
     * \throw bit::io_error if write fails
     */
    void write(FILE *file) const
    {
      fputs("ARRAY1:", file);
      fprintf(file, "%d:", m_bits_per_elem);
      fprintf(file, "%lld:", m_num_elems);
      assert(data_len() == bytes_required(m_num_elems, m_bits_per_elem));
      if (data_len() > 0) {
        size_t ret = fwrite(data(), data_len(), 1, file);
        if (ret != 1)
          throw bit::io_error(
            std::string("bit::Array::write() fwrite failed: ") + 
            strerror(errno));
      }
    }

    /** Read the array from file.  Note that capacity is not stored in
     * the file.  The capacity is set to number of elements stored in
     * the file.
     *
     * \param file = file stream to read from
     * \throw bit::io_error if read fails
     */
    void read(FILE *file)
    {
      int version;
      int ret = fscanf(file, "ARRAY%d:%d:%lld:", 
                       &version, &m_bits_per_elem, &m_num_elems);
      if (ret != 3 || version != 1)
        throw bit::io_error("bit::Array::read() error while reading header");
      m_buffer.resize(bytes_required(m_num_elems, m_bits_per_elem));
      m_capacity = m_num_elems;
      if (data_len() > 0) {
        size_t ret = fread(data(), data_len(), 1, file);
        if (ret != 1) {
          if (ferror(file))
            throw bit::io_error("bit::Array::read() error while reading buffer");
          assert(feof(file));
          throw bit::io_error("bit::Array::read() eof while reading buffer");
        }
      }
    }

  private:

    /** Bits reserved for each element in the array. */
    unsigned int m_bits_per_elem;

    /** Number of elements in the array. */
    u64 m_num_elems;

    /** Number of elements allocated for the underlying buffer.  When
     * values are written outside the buffer with set_grow() method,
     * the size of the underlying buffer is doubled. */
    u64 m_capacity;

    /** Buffer containing the array elements. */
    Buffer m_buffer;

  };

};

#endif /* BIT_ARRAY_HH */
