#ifndef BIT_FLOATARRAY_HH
#define BIT_FLOATARRAY_HH

#include "bit/util.hh"
#include <float.h>
#include "bit/Array.hh"
#include "bit/endian.hh"

namespace bit {
  
  /** Array for storing floats in quantized form.  The floats can be
   * quantized either linearly or with k-means clustering. */
  class FloatArray : public Array {
  public:

    /** Default constructor */
    FloatArray()
      : Array(0, 0), m_mode(mode_default)
    {
    }

    /** Construct an array with initial number of elements. 
     * \param num_elems = number of elements to allocate initially
     * \param bits_per_elem = number of bits to allocate per element
     */
    FloatArray(u64 num_elems, unsigned int bits_per_elem = 0)
      : Array(num_elems, bits_per_elem), m_mode(mode_default)
    {
    }

    /** Copy the mode from another array.
     * \throw bit::invalid_call array was not in default mode before the call
     */
    void set_mode(const FloatArray &array)
    {
      if (m_mode != mode_default)
        throw bit::invalid_call(
          "bit::FloatArray::set_mode() not default mode");
      if (array.m_mode == mode_default)
        return;
      else if (array.m_mode == mode_linear) {
        linear_quantization(array.m_linear_step);
        return;
      }
      assert(false);
    }

    /** Set linear quantization mode and quantize possible elements in
     * the array.
     * \param step = quantization step
     * \throw bit::invalid_argument if step is negative
     * \throw bit::invalid_call if the array is already quantized
     */
    void linear_quantization(double step)
    {
      if (m_mode != mode_default)
	throw bit::invalid_call(
	  "bit::FloatArray::set_linear_quantization() already quantized");
      if (step < 0)
	throw bit::invalid_argument(
	  "bit::FloatArray::set_linear_quantization() negative step");

      if (num_elems() == 0 || bits_per_elem() == 0) {
	m_mode = mode_linear;
	m_linear_step = step;
	return;
      }

      FloatArray new_array(num_elems(), 0);
      new_array.linear_quantization(step);
      for (u64 i = 0; i < num_elems(); i++)
	new_array.set_grow_widen(i, get(i));
      *this = new_array;
    }

    /** Perform linear quantization to given number of bits so that
     * maximum absolute value fits just in.  Does nothing if the array
     * has only zero-elements.
     *
     * \param bits = number of bits to use (remember that the sign
     * required one bit)
     *
     * \throw bit::invalid_argument if bits < 2 or bits > 31
     */
    void linear_quantization_bits(unsigned int bits)
    {
      if (bits < 2 || bits > 31)
	throw bit::invalid_call(
	  "bit::FloatArray::linear_quantization_bits() invalid bits argument");

      if (num_elems() == 0 || bits_per_elem() == 0)
	return;

      float min = FLT_MAX;
      float max = -FLT_MAX;
      for (u64 i = 0; i < num_elems(); i++) {
	float value = get(i);
	if (value < min)
	  min = value;
	if (value > max)
	  max = value;
      }

      float absmax = std::max(util::abs(min), util::abs(max));
      if (absmax == 0)
	return;
      
      linear_quantization(absmax / bit::one_masks[bits - 1]);
    }

    /** Return value of an element.
     * \param elem = the position to access
     * \return the value at the element
     * \throw bit::out_of_range accessing outside the array
     */
    float get(u64 elem) const
    {
      return to_float(Array::get(elem));
    }
     
    /** Set the value of an element.  Note that if the array is
     * already quantized, and the value does not fit well to the
     * existing quantization, the quantization error may be large.
     *
     * \param elem = the position to set
     * \param value = value to store
     * \throw bit::out_of_range accessing outside the array
     */
    void set(u64 elem, float value)
    {
      Array::set(elem, to_uint(value));
    }

    /** Set the value of an element growing the array if necessary.
     * Note that if the array is already quantized, and the value does
     * not fit well to the existing quantization, the quantization
     * error may be large.
     *
     * \param elem = the position to set
     * \param value = value to store
     */
    void set_grow(u64 elem, float value)
    {
      Array::set_grow(elem, to_uint(value));
    }

    /** Set the value of an element growing the array if necessary and
     * widening bit_width if necessary.  Note that if the array is
     * already quantized, and the value does not fit well to the
     * existing quantization, the quantization error may be large.
     *
     * \param elem = the position to set
     * \param value = value to store
     */
    void set_grow_widen(u64 elem, float value)
    {
      Array::set_grow_widen(elem, to_uint(value));
    }

    /** Convert float value to bit representation. 
     * \param value = float to convert
     * \param return the converted bit representation 
     */
    u32 to_uint(float value) const
    {
      switch (m_mode) {
      case mode_default:
	return float2uint(value);
      case mode_linear:
	return quantize_float(value, m_linear_step);
      case mode_kmeans:
	break;
      }
      assert(false);
      return 0;
    }

    /** Convert an integer bit representation back to float
     * \param i = integer to convert
     * \param return the converted float
     */
    float to_float(u32 i) const
    {
      switch (m_mode) {
      case mode_default:
	return uint2float(i);
      case mode_linear:
	return unquantize_float(i, m_linear_step);
      case mode_kmeans:
	break;
      }
      assert(false);
      return 0;
    }

    /** Write the array in file.  
     *
     * \param file = file stream to write to
     * \throw bit::io_error if write fails
     */
    void write(FILE *file) const
    {
      fprintf(file, "FARRAY1:%d:", (int)m_mode);
      size_t ret = endian::write8(m_linear_step, file);
      try {
	if (ret != 1)
	  throw io_error("writing m_linear_step failed");
	Array::write(file);
      }
      catch (std::exception &e) {
	throw io_error(
	  std::string("bit::FloatArray::write() failed: ") + e.what());
      }
    }

    /** Read the array from file. 
     *
     * \param file = file stream to read from
     * \throw bit::io_error if read fails
     */
    void read(FILE *file)
    {
      int version;
      int mode;
      int ret = fscanf(file, "FARRAY%d:%d:", &version, &mode);
      if (ret != 2 || version != 1)
        throw bit::io_error(
	  "bit::FloatArray::read() error while reading header");
      m_mode = (Mode)mode;
      try {
	size_t ret = endian::read8(&m_linear_step, file);
	if (ret != 1)
	  throw io_error("reading m_linear_step failed");
	Array::read(file);
      }
      catch (std::exception &e) {
	throw io_error(
	  std::string("bit::FloatArray::read() failed: ") + e.what());
      }
    }

  private:
    
    /** Quantization mode. */
    enum Mode { mode_default = 0, mode_linear, mode_kmeans } m_mode;

    /** The quantization step for linear quantization (used only in mode_linear). */
    double m_linear_step;

  };

};

#endif /* BIT_FLOATARRAY_HH */
