#ifndef BIT_ALGO_HH
#define BIT_ALGO_HH

#include "bit/exceptions.hh"
#include "bit/base.hh"

namespace bit {

  /** Find the maximum element from an array. */
  template <class A>
  u32 max(const A &array)
  {
    u32 max = 0;
    for (u64 i = 0; i < array.num_elems(); i++) {
      u32 value = array.get(i);
      if (value > max)
        max = value;
    }
    return max;
  }

  /** Find the first index of a value from a sorted array.
   * 
   * \param array = the array to be examined
   * \param value = the value to be searched
   * \param first = the first index in the range considered
   * \param limit = the index limiting the range (not included in the range)
   * \return the index of the value or \c limit if not found
   * \throw bit::invalid_argument if range is negative
   * \throw bit::out_of_range if range exceeds array size
   */
  template <class A>
  u64 binary_search(const A &array, u32 value, u64 first, u64 limit)
  {
     u64 bound = lower_bound(array, value, first, limit);
     if (bound < limit && array.get(bound) == value)
       return bound;
     return limit;
  }

  /** Find the smallest index at which \c value could be inserted
   * maintaining \c array in sorted order.
   * 
   * \param array = the array to be examined
   * \param value = the candidate value
   * \param first = the first index in the range considered
   * \param limit = the index limiting the range (not included in the range)
   * \return the index at which the value should be inserted
   * \throw bit::invalid_argument if range is negative
   * \throw bit::out_of_range if range exceeds array size
   */
  template <class A>
  u64 lower_bound(const A &array, u32 value, u64 first, u64 limit)
  {
    if (first > limit)
      throw bit::invalid_argument(
        "bit::lower_bound(): negative range");
    if (limit > array.num_elems())
      throw bit::out_of_range("bit::lower_bound(): out of range");

    u64 middle;
    u64 half;
    u64 len = limit - first;

    while (len > 5) { // NOTE: magic threshold to do linear search
      half = len / 2;
      middle = first + half;

      // Equal
      u32 cur_value = array.get(middle);

      // First half
      if (cur_value >= value) {
        limit = middle;
        len = limit - first;
      }

      // Second half
      else {
        first = middle + 1;
        len = limit - first;
      }
    }

    while (first < limit) {
      u32 cur_value = array.get(first);
      if (cur_value >= value)
        break;
      first++;
    }

    return first;
  }

  /** From a sorted array, find the largest index that has a
   * equal or smaller value than specified.
   * 
   * \param array = the array to be examined
   * \param value = the candidate value
   *
   * \return the largest index containing equal or smaller value;
   * max_u64 if there exists no such index inside range (value is
   * smaller than first value in the range); 
   */
  template <class A>
  u64 last_leq(const A &array, u32 value)
  {
    return last_leq(array, value, 0, array.num_elems());
  }

  /** From a sorted array range, find the largest index that has a
   * equal or smaller value than specified.
   * 
   * \param array = the array to be examined
   * \param value = the candidate value
   * \param first = the start of the range
   * \param limit = the limit of the range (not included in range)
   *
   * \return the largest index containing equal or smaller value;
   * max_u64 if there exists no such index inside range (value is
   * smaller than first value in the range); 
   *
   * \throw bit::invalid_argument if negative range
   * \throw bit::out_of_range if limit greater than number of elements
   */
  template <class A>
  u64 last_leq(const A &array, u32 value, u64 first, u64 limit)
  {
    if (first > limit)
      throw bit::invalid_argument("bit::last_leq() negative range");
    if (limit > array.num_elems())
      throw bit::out_of_range("bit::last_leq() limit out of range");

    u64 middle;
    u64 half;
    u64 len = limit - first;

    while (len > 5) { // NOTE: magic threshold to do linear search
      half = len / 2;
      middle = first + half;

      // Equal
      u32 cur_value = array.get(middle);

      // First half
      if (cur_value <= value) {
        first = middle;
        len = limit - first;
      }

      // Second half
      else {
        limit = middle;
        len = limit - first;
      }
    }

    u64 orig_first = first;
    while (first < limit) {
      u32 cur_value = array.get(first);
      if (cur_value > value)
        break;
      first++;
    }

    if (first == orig_first)
      return max_u64;

    return first - 1;
  }

  template <class A>
  u64 last_leq_naive(const A &array, u32 value, u64 first, u64 limit)
  {
    if (first > limit)
      throw bit::invalid_argument("bit::last_leq() negative range");
    if (limit > array.num_elems())
      throw bit::out_of_range("bit::last_leq() limit out of range");

    u64 i;
    for (i = first; i < limit; i++) {
      if (array.get(i) > value) {
        break;
      }
    }
    if (i == first)
      return max_u64;
    return i - 1;
  }

};

#endif /* BIT_ALGO_HH */
