#ifndef KNESER_HH
#define KNESER_HH

#include "bit/Array.hh"
#include "bit/Trie.hh"
#include "str/str.hh"
#include "lm/SymbolMap.hh"
#include "util/Progress.hh"
#include "util/util.hh"

namespace bit {

  /** Pruning algorithm for buildling variable-order models with
   * Kneser-Ney smoothing presented in: Reinhard Kneser. Statistical
   * Language Modeling Using a Variable Context Length.  EUROSPEECH
   * 1997.
   *
   * \bug The test suite of Kneser class is not very thoroug.
   */
  class Kneser { 
  public:
  
    typedef Trie<Array> Trie;
    typedef Trie::Iterator Iterator;
    typedef SymbolMap<std::string, int> SymbolMap;
    typedef std::vector<float> FloatVec;
    typedef std::vector<int> IntVec;
    typedef std::vector<int> Ngram;

    /** Default constructor. */
    Kneser()
    {
      m_d1_weight_model = 1;
      m_d1_model = 1;

      m_sentence_start_str = "<s>";
      m_sentence_end_str = "</s>";
      m_sentence_start_id = -1;
      m_sentence_end_id = -1;
      m_progress_skip = 53871;
      reserve_orders(10); // FIXME magic number
    }

    /** Set model used in d1 computation (0 = kn, 1 = abs) */
    void set_d1_model(int model)
    {
      m_d1_model = model;
    }

    /** Set model used as ngram weight in d1 computation (0 = kn, 1 = abs) */
    void set_d1_weight_model(int model)
    {
      m_d1_weight_model = model;
    }

    /** Constant access to symbol map. */ 
    const SymbolMap &symbol_map() const
    {
      return m_symbol_map;
    }

    /** Index of the sentence start symbol. */
    int sentence_start_id() const
    {
      return m_sentence_start_id;
    }

    /** Index of the sentence end symbol. */
    int sentence_end_id() const
    {
      return m_sentence_end_id;
    }

    /** Iterator at the root of the trie. */
    Iterator root() const
    {
      return Iterator(m_trie);
    }

    /** Return the number of ngrams in the unpruned model. */
    u64 num_ngrams() const
    {
      u64 ret = 0;
      for (size_t i = 0; i < m_counts.size(); i++)
        ret += m_counts[i].size();
      return ret;
    }

    /** Return the number of active ngrams in the pruned model. */
    u64 num_active_ngrams() const
    {
      u64 ret = 0;
      for (size_t i = 0; i < m_num_ngrams.size(); i++)
        ret += m_num_ngrams[i];
      return ret;
    }

    /** Find an ngram. 
     * \param vec = ngram to find
     * \return iterator to the ngram, or empty if not found
     */
    template <class T>
    Iterator find(const std::vector<T> &vec) const
    {
      return m_trie.find(vec);
    }

    /** Find an ngram. 
     * \param str = white-space separated list of ngram symbols
     * \return iterator to the ngram, or empty if not found
     */
    Iterator find(const std::string &str) const
    {
      return m_trie.find(ngram(str));
    }

    /** Probability of ngram P(abc) = P(a) P(b|a) P(c|ab) 
     *
     * \param it = ngram iterator for which the probability is computed
     * \return the probability of the ngram
     * \throw bit::invalid_argument if called at root
     */
    float ngram_prob(Iterator it) const
    {
      double ret = 1;
      if (it.length() == 0)
        throw bit::invalid_argument("bit::Kneser::prob_weight() at root");
      while (it.length() > 0) {
        if (it.length() != 1 || it.symbol() != m_sentence_start_id) {
          switch (m_d1_model) {
          case 0:
            ret *= prob_full(it);
            break;
          case 1:
            ret *= prob_abs_full(it);
            break;
          case 2:
            throw bit::invalid_argument(
              "bit::Kneser::ngram_prob() invalid m_d1_model");
          }
        }
        it.goto_parent();
      }
      return ret;
    }

    /** Compute conditional probability P(w|h) using only lower-order
     * probabilities of the beta distribution (with the interpolation
     * weight of h)
     *
     * \throw bit::invalid_argument if \c w is sentence start symbol
     */
    float prob_beta_lower(Iterator it) const
    {
      assert(it.length() > 0);

      double ret = 0;
      double scale = 1;
      u32 symbol = it.symbol();
      it.goto_parent();
      scale *= get_beta_interpolation_numerator(it) /
        get_beta_denominator(it);

      while (!it.is_root()) {
        it.goto_backoff_full();
        if (it.goto_child(symbol)) {
          double numerator = 0;
          if (!is_pruned(it)) {
            numerator = get_beta_numerator(it) - 
              get_beta_discount(it.length() - 1);
            it.goto_parent();
            ret += scale * numerator / get_beta_denominator(it);
          }
          else
            it.goto_parent();
        }

        scale *= get_beta_interpolation_numerator(it) /
          get_beta_denominator(it);
      }

      // Zero-gram probability.  Sentence start is not counted as a
      // real symbol, thus "minus one"
      //
      ret += scale / (m_symbol_map.size() - 1);
      
      return ret;
    }

    /** Convert a string to vector of symbol indices. */
    Ngram ngram(const std::string &str) const
    {
      std::vector<std::string> symbols = str::split(str, " \t", true);
      Ngram ngram(symbols.size());
      for (size_t i = 0; i < symbols.size(); i++)
        ngram[i] = m_symbol_map.index(symbols[i]);
      return ngram;
    }

    /** Compute conditional probability P(w|h) for ngrams found
     * explicitly in the model using the beta distribution.
     *
     * \throw bit::invalid_argument if \c w is sentence start symbol
     */
    float prob_beta_full(const Iterator &it) const
    {
      assert(m_sentence_start_id >= 0);
      assert(it.length() > 0);
      if (it.symbol() == m_sentence_start_id)
        throw bit::invalid_argument(
          "bit::Kneser::prob_beta_full(): called for sentence start");

      unsigned int len = it.length();
      double numerator = 0;
      if (!is_pruned(it))
        numerator = get_beta_numerator(it) - 
          get_beta_discount(len - 1);
      double denominator = get_beta_denominator(it.parent());
      return numerator / denominator + prob_beta_lower(it);
    }

    /** Compute conditional probability P(w|h) for arbitrary ngram
     * using the beta distribution (Warning: implementation may be
     * quite slow).
     *
     * \param vec = std::vector of symbols
     * \throw bit::invalid_argument if w is sentence start symbol or
     * vec is empty
     */
    float prob_beta_full(Ngram ngram) const
    {
      if (ngram.empty())
        throw bit::invalid_argument(
          "bit::Kneser::prob_beta_full(Ngram): ngram empty");
      int symbol = ngram.back();
      if (symbol == m_sentence_start_id)
        throw bit::invalid_argument(
          "bit::Kneser::prob_beta_full(Ngram) sentence start symbol");
        
      double scale = 1;
      double ret = 0;
      while (1) {
        assert(ngram.size() > 0);
        Iterator it = find(ngram);

        // Ngram found?
        if (!it.is_root()) {
          Iterator parent = it.parent();
          double denominator = get_beta_denominator(parent);
          if (!is_pruned(it)) {
            double numerator = get_beta_numerator(it) - 
              get_beta_discount(it.length() - 1);
            assert(numerator >= 0);
            ret += scale * numerator / denominator;
          }
          scale *= get_beta_interpolation_numerator(parent) / denominator;
        }

        // Ngram not found?
        else {
          if (it.length() == 1)
            scale *= get_beta_interpolation_numerator(root()) /
              get_beta_denominator(root());
          else {
            Ngram parent_ngram(ngram.begin(), ngram.end() - 1);
            Iterator parent_it = find(parent_ngram);
            if (!parent_it.is_root())
              scale *= get_beta_interpolation_numerator(parent_it) /
                get_beta_denominator(parent_it);
          }
        }

        if (it.length() == 1)
          break;

        ngram.erase(ngram.begin());
      }

      ret += scale / (m_symbol_map.size() - 1);

      return ret;
    }

    /** Compute conditional probability P(w|h) using only lower-order
     * probabilities (with the interpolation weight of h)
     *
     * \throw bit::invalid_argument if \c w is sentence start symbol
     */
    float prob_lower(Iterator it) const
    {
      assert(it.length() > 0);

      double ret = 0;
      double scale = 1;
      u32 symbol = it.symbol();
      it.goto_parent();
      scale *= interpolation(it);

      while (!it.is_root()) {
        it.goto_backoff_full();
        if (it.goto_child(symbol)) {
          double numerator = 
            sum_nonzero_xg(it) - get_discount(it.length() - 1);
          it.goto_parent();
          double denominator = sum_nonzero_xgx(it);
          ret += scale * numerator / denominator;
        }

        scale *= interpolation(it);
      }

      // Zero-gram probability.  Sentence start is not counted as a
      // real symbol, thus "minus one"
      //
      ret += scale / (m_symbol_map.size() - 1);
      
      return ret;
    }

    /** Compute conditional probability P(w|h).
     * \param it = ngram iterator pointing to (h,w)
     * \param lower_prob = if non-null, the lower-order part "the
     * whole backoff part" of the probability is set here
     * \throw bit::invalid_argument if \c w is sentence start symbol
     */
    float prob_full(const Iterator &it, float *lower_prob = NULL) const
    {
      assert(m_sentence_start_id >= 0);
      assert(it.length() > 0);
      if (it.symbol() == m_sentence_start_id)
        throw bit::invalid_argument(
          "bit::Kneser::prob_full_ikn(): called for sentence start");

      unsigned int len = it.length();
      double numerator = sum_nonzero_xg(it) - get_discount(len - 1);
      double denominator = sum_nonzero_xgx(it.parent());
      double lower = prob_lower(it);
      if (lower_prob != NULL)
        *lower_prob = lower;
      return numerator / denominator + lower;
    }

    /** Compute conditional probability P(w|h) using unmodified counts
     * and using only lower-order probabilities (with the
     * interpolation weight of h)
     *
     * \throw bit::invalid_argument if \c w is sentence start symbol
     */
    float prob_abs_lower(Iterator it) const
    {
      assert(it.length() > 0);

      double ret = 0;
      double scale = 1;
      u32 symbol = it.symbol();
      it.goto_parent();
      scale *= interpolation_abs(it);

      while (!it.is_root()) {
        it.goto_backoff_full();
        if (it.goto_child(symbol)) {
          double numerator = 
            get_count(it) - get_discount(it.length() - 1);
          it.goto_parent();
          double denominator = sum_gx(it);
          ret += scale * numerator / denominator;
        }

        scale *= interpolation_abs(it);
      }

      // Zero-gram probability.  Sentence start is not counted as a
      // real symbol, thus "minus one"
      //
      ret += scale / (m_symbol_map.size() - 1);
      
      return ret;
    }

    /** Compute conditional probability P(w|h) using unmodified counts
     * \param it = ngram iterator pointing to (h,w)
     * \param lower_prob = if non-null, the lower-order part "the
     * whole backoff part" of the probability is set here
     * \throw bit::invalid_argument if \c w is sentence start symbol
     */
    float prob_abs_full(const Iterator &it, float *lower_prob = NULL) const
    {
      assert(m_sentence_start_id >= 0);
      assert(it.length() > 0);
      if (it.symbol() == m_sentence_start_id)
        throw bit::invalid_argument(
          "bit::Kneser::prob_abs_full(): called for sentence start");

      unsigned int len = it.length();
      double numerator = get_count(it) - get_discount(len - 1);
      double denominator = sum_gx(it.parent());
      double lower = prob_abs_lower(it);
      if (lower_prob != NULL)
        *lower_prob = lower;
      return numerator / denominator + lower;
    }

    /** Check if iterator is pruned. */
    bool is_pruned(const Iterator &it) const
    {
      return get_value(m_pruned, it) > 0;
    }

    /** Get ngram count at iterator returning.  It is safe to call
     * this even for iterators for which the count has not been set.
     * Zero is returned in that case.
     *
     * \return the count of the ngram
     */
    u32 get_count(const Iterator &it) const
    {
      return get_value(m_counts, it);
    }

    u32 sum_gx(const Iterator &it) const 
    {
      if (it.is_root())
        return m_sum_gx0;
      return get_value(m_sum_gx, it);
    }

    u32 sum_nonzero_xg(const Iterator &it) const
    {
      return get_value(m_sum_nonzero_xg, it);
    }

    u32 sum_nonzero_gx(const Iterator &it) const
    {
      if (it.is_root())
        return m_sum_nonzero_gx0;
      return get_value(m_sum_nonzero_gx, it);

//       u32 ret = 0;
//       if (it.goto_first_child()) {
//         do {
//           if (get_count(it) > 0)
//             ret++;
//         } while (it.goto_next_sibling());
//       }
//       return ret;
    }

    u32 sum_nonzero_xgx(const Iterator &it) const
    {
      if (it.is_root())
        return m_sum_nonzero_xgx0;
      return get_value(m_sum_nonzero_xgx, it);

//       u32 ret = 0;
//       if (it.goto_first_child()) {
//         do {
//           ret += sum_nonzero_xg(it);
//         } while (it.goto_next_sibling());
//       }
//       return ret;
    }

    float get_beta_numerator(const Iterator &it) const
    {
      if (is_pruned(it))
        return 0;

      int term1 = get_count(it) - get_value(m_sum_xg_not_pruned, it);

      float term2;
      if (1) {
        term2 = get_beta_discount(it.length()) * 
          get_value(m_sum_nonzero_xg_not_pruned, it);
      }
      else {
        term2 = get_value(m_sum_nonzero_xg_not_pruned, it);
        static bool printed = false;
        if (!printed) {
          fprintf(stderr, "\n"
                  "**************************************************\n"
                  "* WARNING: using d = 1 in beta numerator\n"
                  "*\n");
        }
        printed = true;
      }
      
      if (term1 < 0)
        throw std::out_of_range(
          str::fmt(256, "N(h,w) - Sum_v N(v,h,w) negative: %d\n", term1));

      return term1 + term2;
    }

    /** Return the beta denominator (1 for highest-order ngrams) */
    float get_beta_denominator(const Iterator &it) const
    {
      if (it.is_root())
        return m_beta_denominator0;
      if (it.length() > m_beta_denominator.size())
        return 1;
      return get_value(m_beta_denominator, it);
    }

    float get_beta_interpolation_numerator(const Iterator &it) const
    {
      if (it.is_root())
        return m_beta_interpolation_numerator0;
      if (it.length() > m_beta_interpolation_numerator.size())
        return 1;
      return get_value(m_beta_interpolation_numerator, it);
    }

    /** Return the d1 measure of an ngram.
     * \throw bit::invalid_argument if called for unigrams (we never prune unigrams)
     */
    float get_d1(const Iterator &it) const
    {
      if (it.length() < 2)
        throw bit::invalid_argument("bit::Kneser::get_d1(): too low order");
      return get_value(m_d1, it);
    }

    /** Return the unnormalized d2 measure of an ngram.
     * \throw bit::invalid_argument if called for unigrams (we never prune unigrams)
     */
    float get_d2(const Iterator &it) const
    {
      if (it.length() < 2)
        throw bit::invalid_argument("bit::Kneser::get_d2(): too low order");
      return get_value(m_d2, it) / get_value(m_d2_norm, it);
    }

    /** Compute the number of active (not pruned) children. */
    int num_active_children(Iterator it) const
    {
      int ret = 0;
      if (it.goto_first_child()) {
        do {
          if (!is_pruned(it) && it.symbol() != m_sentence_start_id)
            ret++;
        } while (it.goto_next_sibling());
      }
      return ret;
    }

    /** Printable string of a ngram. 
     * \param ngram = ngram to convert
     * \return string representing the ngram
     */
    template <class T>
    std::string 
    ngram_str(const std::vector<T> &ngram) const
    {
      std::string str;
      for (size_t o = 0; o < ngram.size(); o++) {
        if (o != 0)
          str.append(" ");
        str.append(m_symbol_map.at(ngram[o]));
      }
      return str;
    }

    /** Write trie structure, counts, modified counts, sum_nonzero_xg,
     * sum_nonzero_xgx and sum_nonzero_gx in a file. */
    void write_binary_counts(FILE *file) const
    {
      m_symbol_map.write(file);
      m_trie.write(file);
      util::write(file, m_counts);
      util::write(file, m_sum_nonzero_xg);
      util::write(file, m_sum_nonzero_xgx);
      util::write(file, m_sum_nonzero_gx);
      util::write(file, m_sum_gx);
      fwrite(&m_sum_nonzero_xgx0, sizeof(m_sum_nonzero_xgx0), 1, file);
      fwrite(&m_sum_nonzero_gx0, sizeof(m_sum_nonzero_gx0), 1, file);
      fwrite(&m_sum_gx0, sizeof(m_sum_gx0), 1, file);
    }

    /** Read trie structure, counts, modified counts, sum_nonzero_xg,
     * sum_nonzero_xgx and sum_nonzero_gx from a file. */
    void read_binary_counts(FILE *file) 
    {
      m_symbol_map.read(file);
      m_trie.read(file);
      util::read(file, m_counts);
      util::read(file, m_sum_nonzero_xg);
      util::read(file, m_sum_nonzero_xgx);
      util::read(file, m_sum_nonzero_gx);
      util::read(file, m_sum_gx);
      size_t ret = 0;
      ret += fread(&m_sum_nonzero_xgx0, sizeof(m_sum_nonzero_xgx0), 1, file);
      ret += fread(&m_sum_nonzero_gx0, sizeof(m_sum_nonzero_gx0), 1, file);
      ret += fread(&m_sum_gx0, sizeof(m_sum_gx0), 1, file);
      if (ret != 3)
        throw io_error("bit::Kneser::read_binary_counts() failed");

      m_sentence_start_id = m_symbol_map.index(m_sentence_start_str);
      m_sentence_end_id = m_symbol_map.index(m_sentence_end_str);
      m_num_ngrams.resize(m_counts.size());
      for (size_t i = 0; i < m_counts.size(); i++)
        m_num_ngrams.at(i) = m_counts[i].size();
    }

    /** Write d1 and d2. */
    void write_binary_d1d2(FILE *file) const
    {
      util::write(file, m_d1);
      util::write(file, m_d2);
      util::write(file, m_d2_norm);
    }

    /** Read d1 and d2. */
    void read_binary_d1d2(FILE *file) 
    {
      util::read(file, m_d1);
      util::read(file, m_d2);
      util::read(file, m_d2_norm);
    }

    /** Write model in ARPA format. */
    void write_arpa(FILE *file) const 
    {
      fprintf(file, "\\data\\\n");
//      for (size_t l = 0; l < order(); l++)
//        fprintf(file, "ngram %d=%lld\n", (int)l+1, 
//                m_score_arrays.at(l).num_elems());

      for (size_t o = 0; o < m_counts.size(); o++) {
        fprintf(file, "\n\\%zd-grams:\n", o+1);
        Iterator it = root();
        while (it.goto_next_on_level(o)) {

          float log_prob;
          if (it.length() == 1 && it.symbol() == m_sentence_start_id)
            log_prob = -99;
          else
            log_prob = log10(prob_full(it));

          fprintf(file, "%g\t%s", log_prob,
                  ngram_str(it.symbol_vec()).c_str());
          if (it.num_children() > 0)
            fprintf(file, "\t%g\n", log10(interpolation(it)));
          else
            fputs("\n", file);
        }
      }
      fprintf(file, "\n\\end\\\n");
    }

    /** Write model in ARPA format. */
    void write_beta_arpa(FILE *file) const 
    {
      fprintf(file, "\\data\\\n");
//      for (size_t l = 0; l < order(); l++)
//        fprintf(file, "ngram %d=%lld\n", (int)l+1, 
//                m_score_arrays.at(l).num_elems());

      for (size_t i = 0; i < m_num_ngrams.size(); i++) {
        assert(m_num_ngrams[i] >= 0);
        if (m_num_ngrams[i] == 0)
          break;
        fprintf(file, "ngram %d=%d\n", (int)i + 1, m_num_ngrams[i]);
      }

      Progress p(m_progress_skip, num_ngrams());
      p.set_report_string("writing arpa:");

      for (size_t o = 0; o < m_counts.size(); o++) {
        fprintf(file, "\n\\%zd-grams:\n", o+1);
        Iterator it = root();
        while (it.goto_next_on_level(o)) {
          p.step();
          if (get_value(m_pruned, it) > 0)
            continue;

          float log_prob;
          if (it.length() == 1 && it.symbol() == m_sentence_start_id)
            log_prob = -99;
          else
            log_prob = log10(prob_beta_full(it));

          fprintf(file, "%g\t%s", log_prob,
                  ngram_str(it.symbol_vec()).c_str());
          if (num_active_children(it) > 0)
            fprintf(file, "\t%g\n", 
                    log10(get_beta_interpolation_numerator(it) /
                          get_beta_denominator(it)));
          else
            fputs("\n", file);
        }
      }
      fprintf(file, "\n\\end\\\n");
      p.finish();
    }

    /** Reserve space for orders to avoid reallocing. 
     * \param orders = number of orders used
     */
    void reserve_orders(unsigned int orders)
    {
      m_counts.reserve(orders);
      m_sum_nonzero_xg.reserve(orders);
      m_sum_nonzero_gx.reserve(orders);
      m_sum_nonzero_xgx.reserve(orders);
      m_pruned.reserve(orders);
      m_d1.reserve(orders);
      m_d2.reserve(orders);
      m_d2_norm.reserve(orders);
      m_sum_xg_not_pruned.reserve(orders);
      m_sum_nonzero_xg_not_pruned.reserve(orders);
      m_beta_denominator.reserve(orders);
      m_beta_interpolation_numerator.reserve(orders);
      m_trie.reserve_levels(orders);
    }
    

    /** Read counts from an ASCII file.  Each line contains a white
     * space separated list of strings; the ngram symbols followed by
     * the count.
     *
     * \param file = file stream to read from
     * \param integer_symbols = the ngram symbols are integers
     * \throw bit::io_error if reading fails
     */
    void read_counts(FILE *file, bool integer_symbols = false)
    {
      std::string line;
      std::vector<int> ngram;
      std::vector<std::string> fields;
      int count;
      Progress p(m_progress_skip);
      p.set_report_string("reading counts:");
      while (str::read_line(line, file, true)) {
        try {
          if (integer_symbols) {
            ngram = str::long_vec<int>(line);
            if (ngram.size() < 2)
              throw std::exception();
            count = ngram.back();
            ngram.pop_back();
          }
          else {
            fields = str::split(line, " \t", true);
            if (fields.size() < 2)
              throw std::exception();
            count = str::str2long(fields.back());
            fields.pop_back();
            ngram.resize(fields.size());
            for (size_t i = 0; i < fields.size(); i++)
              ngram[i] = m_symbol_map.insert(fields[i]);
          }
        }
        catch (std::exception &e) {
          throw bit::io_error(
            std::string("bit::Kneser::read_counts(): invalid line: ") + line);
        }
        
        add(ngram, count);
        p.step();
      }
      p.finish();

      m_num_ngrams.resize(m_counts.size());
      for (size_t i = 0; i < m_counts.size(); i++)
        m_num_ngrams.at(i) = m_counts[i].size();
    }

    /** Compute Kneser-Ney modified counts.  The modified count stored
     * at ngram (a,b,c) is the number of ngrams (x,a,b,c) which have
     * positive count.  If no such ngrams exist, the original count is
     * used instead.
     */
    void compute_sums()
    {
      assert(m_sentence_start_id < 0);
      m_sentence_start_id = m_symbol_map.index(m_sentence_start_str);
      m_sentence_end_id = m_symbol_map.index(m_sentence_end_str);
      {
        Iterator it = root();
        it.goto_child(m_sentence_start_id);
        set_count(it, 0);
        fprintf(stderr, "WARNING: setting count(%s) = 0\n", 
                m_sentence_start_str.c_str());
      }

      assert(m_sum_nonzero_xg.empty());
      assert(m_sum_nonzero_xgx.empty());
      m_sum_gx.resize(m_counts.size() - 1);
      m_sum_nonzero_xg.resize(m_counts.size());
      m_sum_nonzero_xgx.resize(m_counts.size() - 1);
      m_sum_nonzero_gx.resize(m_counts.size() - 1);
      for (size_t i = 0; i < m_sum_nonzero_xg.size(); i++)
        m_sum_nonzero_xg[i].resize(m_counts[i].size());
      for (size_t i = 0; i < m_sum_nonzero_xgx.size(); i++) {
        m_sum_gx[i].resize(m_counts[i].size());
        m_sum_nonzero_xgx[i].resize(m_counts[i].size());
        m_sum_nonzero_gx[i].resize(m_counts[i].size());
      }
      m_sum_nonzero_xgx0 = 0;
      m_sum_nonzero_gx0 = 0;
      m_sum_gx0 = 0;

      // Compute modified counts
      //
      {
        Progress p(m_progress_skip, num_ngrams());
        p.set_report_string("modified counts:");
        Iterator it = root();
        while (it.goto_next_depth_first()) {
          p.step();

          // sum_gx is needed only if absolute discounting is used
          if (m_d1_model == 1 || m_d1_weight_model == 1) {
            Iterator parent(it.parent());
            u32 count = get_count(it);
            if (parent.length() == 0)
              m_sum_gx0 += count;
            else
              add_value(m_sum_gx, parent, count);
          }

          if (it.length() < 2)
            continue;
        
          Iterator bo_it(it);
          if (bo_it.goto_backoff_once())
            add_value(m_sum_nonzero_xg, bo_it, 1);
        }
        p.finish();
      }

      // Use original counts for ngrams that do not have contexts
      // (ngrams starting with <s> for example)
      //
      {
        Progress p(m_progress_skip, num_ngrams());
        p.set_report_string("without context:");
        for (size_t o = 0; o < m_counts.size(); o++) {
          IntVec &src_array = m_counts.at(o);
          IntVec &sum_nonzero_xg = m_sum_nonzero_xg.at(o);
          for (u64 i = 0; i < src_array.size(); i++) {
            p.step();
            if (sum_nonzero_xg.at(i) == 0)
              sum_nonzero_xg[i] = src_array.at(i);
          }
        }
        p.finish();
      }

      // Compute sum_nonzero_xgx and sum_nonzero_gx
      //
      {
        Progress p(m_progress_skip, num_ngrams());
        p.set_report_string("modified sum counts:");
        Iterator it = root();
        while (it.goto_next_depth_first()) {
          p.step();
          if (it.symbol() == m_sentence_start_id)
            continue;
          u32 value = sum_nonzero_xg(it);
          Iterator parent = it.parent();
          if (parent.is_root()) {
            m_sum_nonzero_xgx0 += value;
            m_sum_nonzero_gx0++;
          }
          else {
            add_value(m_sum_nonzero_xgx, parent, value);
            add_value(m_sum_nonzero_gx, parent, 1);
          }
        }
        p.finish();
      }
    }

    /** Compute d1 measure for all ngrams (n > 1) using log10. */
    void compute_d1()
    {
      float max_d1 = -1e30;
      float min_d1 = 1e30;

      assert(m_d1.empty());
      m_d1.resize(m_counts.size());
      for (size_t i = 0; i < m_counts.size(); i++)
        m_d1.at(i).resize(m_counts.at(i).size());

      Iterator it = root();

      Progress p(m_progress_skip, num_ngrams());
      p.set_report_string("computing d1:");
      while (it.goto_next_depth_first()) {
        p.step();
        if (it.length() < 2)
          continue;

        double orig = 0;
        float lower_prob = 0;
        if (m_d1_model == 0) {
          orig = prob_full(it, &lower_prob);
        }
        else if (m_d1_model == 1) {
          orig = prob_abs_full(it, &lower_prob);
        }
        else
          throw bit::invalid_argument(
            "bit::Kneser::compute_d1() invalid m_d1_model");

        double d1 = ngram_prob(it) * log10(orig / lower_prob);

        if (0) {
          fprintf(stderr, "d1: %12g %s\n", d1, 
                  ngram_str(it.symbol_vec()).c_str());
        }

        if (!(d1 > 0 && d1 < 1e10)) {
          fprintf(stderr, "WARNING: d1 = %g for %s\n", d1, 
                  ngram_str(it.symbol_vec()).c_str());
        }

        set_value(m_d1, it, d1);
        if (d1 < min_d1)
          min_d1 = d1;
        if (d1 > max_d1)
          max_d1 = d1;
      }
      p.finish();

      fprintf(stderr, "min_d1 = %g\n", min_d1);
      fprintf(stderr, "max_d1 = %g\n", max_d1);
    }

    /** Compute d2 measure for each node when no nodes are yet pruned.
     * Method prune_threshold modifies d2 when pruning nodes. */
    void compute_d2_full()
    {
      assert(m_d2.empty());
      assert(!m_d1.empty());
      assert(m_d2_norm.empty());
      m_d2.resize(m_counts.size());
      m_d2_norm.resize(m_counts.size());
      for (size_t i = 0; i < m_counts.size(); i++) {
        m_d2.at(i).resize(m_counts.at(i).size());
        m_d2_norm.at(i).resize(m_counts.at(i).size());
      }

      // Compute d2 measures bottom-up fashion
      //
      Iterator it = root();
      Progress p(m_progress_skip, num_ngrams());
      p.set_report_string("full d2:");
      while (it.goto_next_depth_first_post()) {
        p.step();
        if (it.length() < 2)
          continue;

        D2Norm pair(get_d1(it), 1);
        Iterator child_it(it);
        if (child_it.goto_first_child()) {
          do {
            float d2 = get_value(m_d2, child_it);
            u32 norm = get_value(m_d2_norm, child_it);
            pair.add(d2, norm);
          } while (child_it.goto_next_sibling());
        }

        set_value(m_d2, it, pair.d2);
        set_value(m_d2_norm, it, pair.norm);

        if (0) {
          fprintf(stderr, "d2: %12g %s\n", pair.d2 / pair.norm, 
                  ngram_str(it.symbol_vec()).c_str());
        }
      }
      p.finish();
    }

    /** Compute d2 measure for all ngrams (n > 1).  For each ngram we
     * can precompute what the d2 measure will be at the time when the
     * node would be pruned.  We can do this in a bottom-up
     * fashion. */
    void compute_d2_trick()
    {
      fprintf(stderr, "WARNING: using erroneous d2 measure\n");

      assert(m_d2.empty());
      assert(!m_d1.empty());
      assert(m_d2_norm.empty());
      m_d2.resize(m_counts.size());
      m_d2_norm.resize(m_counts.size());
      for (size_t i = 0; i < m_counts.size(); i++)
        m_d2.at(i).resize(m_counts.at(i).size());

      // Compute d2 measures bottom-up fashion
      //
      std::vector<D2Norm> child_pairs;
      Iterator it = root();
      Progress p(m_progress_skip, num_ngrams());
      p.set_report_string("computing d2:");
      while (it.goto_next_depth_first_post()) {
        p.step();
        if (it.length() < 2)
          continue;

        D2Norm pair(get_d1(it), 1);
        Iterator child_it(it);
        child_pairs.clear();
        if (child_it.goto_first_child()) {
          do {
            float d2 = get_value(m_d2, child_it);
            u32 norm = get_value(m_d2_norm, child_it);
            pair.add(d2, norm);
            child_pairs.push_back(D2Norm(d2, norm));
          } while (child_it.goto_next_sibling());
          std::sort(child_pairs.begin(), child_pairs.end());
        }

        // Virtually remove childrens until the parent's d2 is smaller
        // than that of best children.
        //
        for (size_t i = 0; i < child_pairs.size(); i++) {
          if (pair < child_pairs[i])
            break;
          pair.add(-child_pairs[i].d2, -child_pairs[i].norm);
        }

        set_value(m_d2, it, pair.d2);
        set_value(m_d2_norm, it, pair.norm);
      }
      p.finish();
    }

    /** Prune ngram, modify parents' d2 measure, and remove
     * children. 
     *
     * \throw bit::invalid_argument if pruned already or ngram is
     * shorter than 2-gram
     */
    void prune_ngram(Iterator it)
    {
      if (it.length() < 2)
        throw bit::invalid_argument(
          "bit::Kneser::prune_ngram() ngram shorter than 2-gram");
        
      if (get_value(m_pruned, it) > 0)
        throw bit::invalid_argument(
          "bit::Kneser::prune_ngram() ngram pruned already");
      set_value(m_pruned, it, 1);
      m_num_ngrams.at(it.length() - 1)--;
      
      float d2 = get_value(m_d2, it);
      int d2_norm = get_value(m_d2_norm, it);

      if (0) {
        fprintf(stderr, "pruned %12g %12g / %d %s\n", 
                get_value(m_d1, it), 
                d2, d2_norm, 
                ngram_str(it.symbol_vec()).c_str());
      }

      // Mark children pruned too
      {
        Iterator child_it(it);
        unsigned int len = it.length();
        while (child_it.goto_next_depth_first()) {
          if (child_it.length() <= len)
            break;
          if (get_value(m_pruned, child_it) > 0)
            continue;
          if (0) {
            fprintf(stderr, "pruned %12g %12g / %d %s\n", 
                    get_value(m_d1, child_it),
                    get_value(m_d2, child_it),
                    get_value(m_d2_norm, child_it),
                    ngram_str(child_it.symbol_vec()).c_str());
          }
          set_value(m_pruned, child_it, 1);
          m_num_ngrams.at(child_it.length() - 1)--;
        }
      }

      // Modify parents' d2 measure
      while (1) {
        it.goto_parent();
        if (it.length() == 1)
          break;
        add_value(m_d2, it, -d2);
        add_value(m_d2_norm, it, -d2_norm);
      }
    }

    /** Prune ngrams (n > 1) whose d2 is under threshold.  Children
     * are processed before parents, and parents' d2 is modified if
     * children are pruned. */
    void prune_threshold(float threshold)
    {
      assert(!m_d1.empty());
      assert(!m_d2.empty());
      assert(m_pruned.empty());
      m_pruned.resize(m_counts.size());

      Progress p(m_progress_skip, num_ngrams());
      p.set_report_string("pruning:");
      Iterator it = root();
      while (it.goto_next_depth_first_post()) {
        p.step();
        if (it.length() < 2)
          continue;
        if (get_d2(it) < threshold)
          prune_ngram(it);
      }
      p.finish();
      fprintf(stderr, "%lld ngrams left\n", num_active_ngrams());
    }

    /** Prune ngrams (n > 1) according to the d2 measure.  The ngrams
     * are not actually removed, but only marked pruned.
     *
     * \param ngrams = number of ngrams to prune
     *
     * \throw bit::invalid_argument if \c ngrams is greater than
     * number of ngrams for (n > 2).
     */
    void prune(unsigned int ngrams)
    {
      assert(!m_d1.empty());
      assert(!m_d2.empty());
      assert(m_pruned.empty());
      m_pruned.resize(m_counts.size());

      size_t num_ngrams = 0;
      for (size_t i = 0; i < m_counts.size(); i++)
        num_ngrams += m_counts.at(i).size();
      std::vector<OrderIndex> vec;
      vec.reserve(num_ngrams);

      Iterator it = root();
      while (it.goto_next_depth_first_post()) {
        assert(it.length() > 0);
        if (it.length() < 2)
          continue;
        vec.push_back(OrderIndex(it.length() - 1, it.symbol_index()));
      }

      if (ngrams > vec.size())
        throw bit::invalid_argument(
          "bit::Kneser::prune() trying to prune too many ngrams");

      Progress p(0, 2);
      p.set_report_string("sorting:");
      p.step();
      std::partial_sort(vec.begin(), vec.begin() + ngrams, vec.end(), 
                        PruneCompare(this));
      p.step();
      p.finish();

      for (size_t i = 0; i < ngrams; i++) {
        m_num_ngrams.at(vec[i].order)--;
        m_pruned.at(vec[i].order).set_grow_widen(vec[i].index, 1);
      }

//       fprintf(stderr, "DEBUG: pruned\n");
//       it = root();
//       while (it.goto_next_depth_first())
//         if (get_value(m_pruned, it) > 0)
//           fprintf(stderr, "  %10g %s\n", get_d2(it), 
//                   ngram_str(it.symbol_vec()).c_str());

      // Prune children of the pruned ngrams if they are not pruned already
      //
      it = root();
      while (it.goto_next_depth_first()) {
        if (it.length() < 2)
          continue;
        if (get_value(m_pruned, it.parent()) > 0 &&
            get_value(m_pruned, it) == 0)
        {
          unsigned int order = it.length() - 1;
          m_num_ngrams.at(order)--;
          set_value(m_pruned, it, 1);
//           fprintf(stderr, 
//                   "WARNING: pruning \"%s\" because parent was pruned\n",
//                   ngram_str(it.symbol_vec()).c_str());
        }
      }

    }

    /** Precompute numerator terms for Kneser's improved back-off
     * distribution.  Note that discounting weights are modified too.
     */
    void compute_beta_numerator_terms()
    {
      assert(!m_discounts.empty());
      m_beta_discounts = m_discounts;
      for (size_t i = m_beta_discounts.size() - 1; i > 0; i--)
        m_beta_discounts[i - 1] *= m_beta_discounts[i];

      assert(m_sentence_start_id >= 0);
      assert(m_sum_xg_not_pruned.empty());
      assert(m_sum_nonzero_xg_not_pruned.empty());
      m_sum_xg_not_pruned.resize(m_counts.size() - 1);
      m_sum_nonzero_xg_not_pruned.resize(m_counts.size() - 1);
      for (size_t i = 0; i < m_counts.size() - 1; i++) {
        m_sum_xg_not_pruned.at(i).resize(m_counts.at(i).size());
        m_sum_nonzero_xg_not_pruned.at(i).resize(m_counts.at(i).size());
      }

      Iterator it = root();
      Progress p(m_progress_skip, num_ngrams());
      p.set_report_string("beta numerator terms:");
      while (it.goto_next_depth_first()) {
        p.step();
        if (get_value(m_pruned, it) > 0)
          continue;
        if (it.length() < 2)
          continue;

        Iterator bo_it(it);
        if (bo_it.goto_backoff_once()) {
          add_value(m_sum_xg_not_pruned, bo_it, get_count(it));
          add_value(m_sum_nonzero_xg_not_pruned, bo_it, 1);
        }
      }
      p.finish();
    }

    /** Precompute denominators for Kneser's improved back-off
     * distribution.  
     */
    void compute_beta_denominator()
    {
      assert(m_beta_denominator.empty());
      m_beta_denominator.resize(m_counts.size() - 1);
      m_beta_denominator0 = 0;
      for (size_t i = 0; i < m_beta_denominator.size(); i++)
        m_beta_denominator.at(i).resize(m_counts.at(i).size());

      Iterator it = root();
      Progress p(m_progress_skip, num_ngrams());
      p.set_report_string("beta denominator:");
      while (it.goto_next_depth_first()) {
        p.step();

        if (it.length() <= m_beta_denominator.size()) {
          int pruned_counts;
          compute_active_children(it, &pruned_counts);
          set_value(m_beta_denominator, it, pruned_counts);
        }

        if (it.symbol() == m_sentence_start_id)
          continue;

        Iterator parent_it(it);
        parent_it.goto_parent();

//        fprintf(stderr, "%10g %s\n", get_beta_numerator(it), 
//                ngram_str(it.symbol_vec()).c_str());

        float numerator = get_beta_numerator(it);
        if (parent_it.is_root())
          m_beta_denominator0 += numerator;
        else
          add_value(m_beta_denominator, parent_it, numerator);
      }
      p.finish();
    }

    /** Compute number of active children and sum of pruned child counts. */
    int compute_active_children(Iterator it, int *pruned_counts = NULL)
    {
      int num_active_children = 0;
      if (pruned_counts != NULL)
        (*pruned_counts) = 0;
      if (it.goto_first_child()) {
        do {
          if (it.symbol() == m_sentence_start_id)
            continue;
          if (is_pruned(it)) {
            if (pruned_counts != NULL)
              (*pruned_counts) += get_count(it);
            continue;
          }
              
          num_active_children++;
        } while (it.goto_next_sibling());
      }
      return num_active_children;
    }

    /** Compute beta distribution and interpolation.  
     *
     * \throw bit::out_of_range if a numerator was less than discount
     */
    void compute_beta_interpolation_numerator()
    {
      assert(m_beta_interpolation_numerator.empty());
      m_beta_interpolation_numerator.resize(m_counts.size() - 1);
      for (size_t i = 0; i < m_beta_interpolation_numerator.size(); i++)
        m_beta_interpolation_numerator.at(i).resize(m_counts.at(i).size());

      m_beta_interpolation_numerator0 = 0;
      Progress p(m_progress_skip, num_ngrams());
      p.set_report_string("beta interpolation numerator:");
      Iterator it = root();
      do {
        p.step();

        if (it.length() > m_beta_interpolation_numerator.size())
          continue;

        if (!it.is_root()) {
          if (is_pruned(it)) {
            set_value(m_beta_interpolation_numerator, it, 1);
            continue;
          }
        }

        int pruned_counts = 0;
        int num_active_children = compute_active_children(it, &pruned_counts);
        double interpolation = get_beta_discount(it.length()) *
          num_active_children + pruned_counts;

        if (it.is_root())
          m_beta_interpolation_numerator0 = interpolation;
        else
          set_value(m_beta_interpolation_numerator, it, interpolation);
      } while (it.goto_next_depth_first());
      p.finish();
    }

    /** Increment count
     * \param vec = ngram whose count is incremented
     * \param value = the amount of increment
     * \return iterator to the ngram
     * \throw bit::invalid argument if ngram's parent was not in the model
     */
    Iterator add(const std::vector<int> &vec, int value)
    {
      Iterator it = m_trie.insert(vec);
      add_value(m_counts, it, value);
      return it;
    }

    /** Set the discounting parameter.
     *
     * \param order = order for which the discount is set.  Zero is
     * the discount for unigrams.
     *
     * \param value = value of the discount
     */
    void set_discount(unsigned int order, float value)
    {
      m_discounts.resize(order + 1);
      m_discounts.at(order) = value;
    }

    /** Get the discount parameters.
     * \param order = order for which the discount is fetched (0 is unigrams)
     * \return the discount parameter
     * \throw bit::invalid_call if discount not set
     */
    float get_discount(unsigned int order) const
    {
      if (m_discounts.empty())
        throw bit::invalid_call(
          "bit::Kneser::get_discount() discount not set");
      if (order >= m_discounts.size())
        return m_discounts.back();
      return m_discounts.at(order);
    }

    /** Get the beta discount parameters.
     * \param order = order for which the discount is fetched (0 is unigrams)
     * \return the discount parameter
     * \throw bit::invalid_call if discount not set
     */
    float get_beta_discount(unsigned int order) const
    {
      if (m_beta_discounts.empty())
        throw bit::invalid_call(
          "bit::Kneser::get_beta_discount() discount not set");
      if (order >= m_beta_discounts.size())
        return m_beta_discounts.back();
      return m_beta_discounts.at(order);
    }

    float interpolation(const Iterator &it) const
    {
      assert(it.length() == 0 || it.symbol() != m_sentence_end_id);
      double nominator = sum_nonzero_gx(it) * get_discount(it.length());
      double denominator = sum_nonzero_xgx(it);
      return nominator / denominator;
    }

    float interpolation_abs(const Iterator &it) const
    {
      assert(it.length() == 0 || it.symbol() != m_sentence_end_id);
      double nominator = sum_nonzero_gx(it) * get_discount(it.length());
      double denominator = sum_gx(it);
      return nominator / denominator;
    }

    /** Set the count of the ngram at iterator returning.
     *
     * \return the count of the ngram
     */
    void set_count(const Iterator &it, u32 value)
    {
      set_value(m_counts, it, value);
    }

    std::string debug_sum_nonzero_xg_str()
    {
      std::string str;
      Iterator it = root();
      while (it.goto_next_depth_first()) {
        str.append(ngram_str(it.symbol_vec()));
        str.append("\t");
        str.append(str::fmt(64, "%d\n", sum_nonzero_xg(it)));
      }
      return str;
    }

    void debug_write_counts(FILE *file)
    {
      fprintf(file, "sum_nonzero_xg_not_pruned:\n");
      Iterator it = root();
      while (it.goto_next_depth_first()) {
        fprintf(file, "%s\t  bdenom=%.2f  bnum=%.2f\n", 
                ngram_str(it.symbol_vec()).c_str(),
                it.length() < 3 ? get_beta_denominator(it) : -1,
                get_beta_numerator(it)
          );
                
      }
    }


  private:

    float get_value(const std::vector<FloatVec> &arrays, const Iterator &it) const
    {
      unsigned int len = it.length();
      u32 index = it.symbol_index();
      return arrays.at(len - 1).at(index);
    }

    int get_value(const std::vector<IntVec> &arrays, const Iterator &it) const
    {
      unsigned int len = it.length();
      if (len > arrays.size())
        return 0;
      u32 index = it.symbol_index();
      const IntVec &vec = arrays.at(len - 1);
      if (index >= vec.size())
        return 0;
      return vec[index];
    }

    u32 get_value(const std::vector<Array> &arrays, const Iterator &it) const
    {
      unsigned int len = it.length();
      u32 index = it.symbol_index();
      if (len > arrays.size())
        return 0;
      const Array &array = arrays.at(len - 1);
      if (index >= array.num_elems())
        return 0;
      return array.get(index);
    }

    void set_value(std::vector<FloatVec> &arrays, const Iterator &it, 
                   float value)
    {
      unsigned int len = it.length();
      u32 index = it.symbol_index();
      arrays.at(len - 1).at(index) = value;
    }

    void set_value(std::vector<IntVec> &arrays, const Iterator &it, 
                   int value)
    {
      unsigned int len = it.length();
      if (len > arrays.size())
        arrays.resize(len);
      u32 index = it.symbol_index();
      IntVec &vec = arrays.at(len - 1);
      while (vec.size() <= index)
        vec.push_back(0);
      vec[index] = value;
    }

    void set_value(std::vector<Array> &arrays, const Iterator &it, u32 value)
    {
      unsigned int len = it.length();
      u32 index = it.symbol_index();
      if (len > arrays.size())
        arrays.resize(len);
      arrays.at(len - 1).set_grow_widen(index, value);
    }

    void add_value(std::vector<Array> &arrays, const Iterator &it, u32 value)
    {
      set_value(arrays, it, get_value(arrays, it) + value);
    }

    void add_value(std::vector<FloatVec> &arrays, const Iterator &it, 
                   float value)
    {
      set_value(arrays, it, get_value(arrays, it) + value);
    }

    void add_value(std::vector<IntVec> &arrays, const Iterator &it, 
                   int value)
    {
      set_value(arrays, it, get_value(arrays, it) + value);
    }

    void sub_value(std::vector<Array> &arrays, const Iterator &it, u32 value)
    {
      u32 old_value = get_value(arrays, it);
      if (old_value < value)
        throw bit::invalid_argument("bit::Kneser::sub_value() underflow");
      set_value(arrays, it, get_value(arrays, it) - value);
    }

    /** \name Configurable options */
    //@{

    /** Probability distribution used for computing ngram weight in d1
     * computation (0 = kneser-ney, 1 = absolute-discounting, 2 =
     * counts). */
    int m_d1_weight_model;

    /** Probability distribution used in d1 computation (0 =
     * kneser-ney, 1 = absolute-discounting). */
    int m_d1_model;

    //@}

    /** Skip between progress reports */
    int m_progress_skip;
    
    /** Trie containing the ngrams. */
    Trie m_trie;

    /** Number of ngrams in the model (updated in pruning) */
    std::vector<int> m_num_ngrams;

    /** Real counts for each order. */
    std::vector<IntVec> m_counts;

    /** Sum of counts */
    std::vector<IntVec> m_sum_gx;

    /** Sum of counts for unigrams. */
    int m_sum_gx0;

    /** Modified counts for each order. */
    std::vector<IntVec> m_sum_nonzero_xg;

    /** Modified normalization counts for each order. */
    std::vector<IntVec> m_sum_nonzero_gx;

    /** Modified normalization counts for unigrams. */
    int m_sum_nonzero_gx0;

    /** Precomputed sum of modified counts for each order. */
    std::vector<IntVec> m_sum_nonzero_xgx;

    /** Precomputed sum of modified counts for unigrams. */
    int m_sum_nonzero_xgx0;

    /** Discount parameters for each order (index 0 is for unigrams) */
    std::vector<float> m_discounts;

    /** Beta discount parameters for each order (index 0 is for
     * unigrams).  In beta distribution the lower order discounts
     * should be multiplied by the higher order discounts. */
    std::vector<float> m_beta_discounts;

    /** Symbol set for mapping between symbols and integers. */
    SymbolMap m_symbol_map;

    /** String of the sentence start symbol. */
    std::string m_sentence_start_str;
        
    /** String of the sentence end symbol. */
    std::string m_sentence_end_str;
        
    /** Index of the sentence start symbol. */
    int m_sentence_start_id;

    /** Index of the sentence end symbol. */
    int m_sentence_end_id;

    /** \name Variables and types used in pruning */
    //@{

    std::vector<Array> m_pruned; 
    std::vector<FloatVec> m_d1;
    std::vector<FloatVec> m_d2;
    std::vector<IntVec> m_d2_norm;
    std::vector<IntVec> m_sum_xg_not_pruned;
    std::vector<IntVec> m_sum_nonzero_xg_not_pruned;

    /** Precomputed denominator for Kneser's improved back-off
     * distribution. */
    std::vector<FloatVec> m_beta_denominator;

    /** Precomputed unigram-denominator for Kneser's improved back-off
     * distribution. */
    float m_beta_denominator0;

    /** Precomputed interpolation weights for Kneser's improved
     * back-off distribution. */
    std::vector<FloatVec> m_beta_interpolation_numerator;

    /** Precomputed interpolation weight for zero-grams. */
    float m_beta_interpolation_numerator0;

    struct OrderIndex {
      OrderIndex(unsigned int order = 0, u32 index = 0) 
        : order(order), index(index) { }
      unsigned int order;
      u32 index;
    };

    struct PruneCompare {
      const Kneser *k;

      PruneCompare(const Kneser *k)
        : k(k)
      {
      }

      bool operator()(const OrderIndex &a, const OrderIndex &b) const
      {
        float d2_a = k->m_d2.at(a.order).at(a.index);
        int d2_a_norm = k->m_d2_norm.at(a.order).at(a.index);
        float d2_b = k->m_d2.at(b.order).at(b.index);
        int d2_b_norm = k->m_d2_norm.at(b.order).at(b.index);
        return (d2_a / d2_a_norm) < (d2_b / d2_b_norm);
      }
    };

    struct D2Norm {
      float d2;
      int norm;
      float value;

      D2Norm() : d2(0), norm(0), value(0)
      { 
      }

      D2Norm(float d2, int norm) : d2(d2), norm(norm), value(d2 / norm)
      { 
      }

      void add(float d2, int norm)
      {
        this->d2 += d2;
        this->norm += norm;
        value = this->d2 / this->norm;
      }

      bool operator<(const D2Norm &p) const
      {
        return value < p.value;
      }
    };

    //@}
  };

}

#endif /* KNESER_HH */
