#include "LM.hh"
#include "ArpaReader.hh"

namespace bit {

  void
  LM::read_arpa(FILE *file, const std::string &sentence_start_str,
                const std::string &sentence_end_str, bool verbose)
  {
    assert(m_start_symbol == -1);
    assert(m_end_symbol == -1);

    ArpaReader reader(file);
    reader.opt.show_progress = verbose;
    reader.read_header();

    // We have to read all 1-grams first to find out sentence start and
    // end symbols.
    //
    try {
      reader.read_order_ngrams(false);
      m_symbol_map = reader.symbol_map;
    }
    catch (std::exception &e) {
      throw io_error(
        std::string("lm::LM::read_arpa(): error while reading unigrams: ") 
        + e.what());
    }

    try {
      m_start_symbol = reader.symbol_map.index(sentence_start_str);
      m_end_symbol = reader.symbol_map.index(sentence_end_str);
    }
    catch (std::string &str) {
      throw invalid_argument(
        std::string("lm::LM::read_arpa(): sentence start '") + 
        sentence_start_str + "' or sentence end '" + sentence_end_str + 
        "' not in unigrams");
    }

    for (int i = 0; i < (int)reader.order_ngrams.size(); i++) {
      ArpaReader::Ngram &ngram = reader.order_ngrams[i];
      assert(ngram.symbols.size() == 1);
      if (ngram.symbols[0] == m_start_symbol) {
        fprintf(stderr, "WARNING: setting probability of sentence start %s "
                "from %g to zero\n", sentence_start_str.c_str(), 
                ngram.log_prob);
        ngram.log_prob = 0;
      }
      insert_ngram(ngram.symbols, ngram.log_prob, ngram.backoff);
    }

    // Continue from 2-grams 
    //
    while (reader.read_order_ngrams(true)) {
      for (int i = 0; i < (int)reader.sorted_order.size(); i++) {
        ArpaReader::Ngram &ngram = reader.order_ngrams[reader.sorted_order[i]];
        insert_ngram(ngram.symbols, ngram.log_prob, ngram.backoff);
      }
    }
  }
  
  void
  LM::write_arpa(FILE *file) const
  {
    fprintf(file, "\\data\\\n");
    for (size_t l = 0; l < order(); l++)
      fprintf(file, "ngram %d=%lld\n", (int)l+1, 
	      m_score_arrays.at(l).num_elems());

    for (size_t l = 0; l < order(); l++) {
      fprintf(file, "\n\\%d-grams:\n", l+1);
      Iterator it = root();
      while (it.goto_next_on_level(l)) {
	fprintf(file, "%g\t%s", score(it), ngram_str(it.symbol_vec()).c_str());
	if (it.num_children() > 0)
	  fprintf(file, "\t%g\n", backoff(it));
	else
	  fputs("\n", file);
      }
    }
    fprintf(file, "\n\\end\\\n");
  }

  void
  LM::write(FILE *file) const
  {
    try {
      assert(m_score_arrays.size() == m_backoff_arrays.size());
      fprintf(file, "LM2:%d:%d:%d:", 
              (int)m_score_arrays.size(), 
              m_start_symbol, m_end_symbol);
      m_symbol_map.write(file);
      m_trie.write(file);
      for (size_t i = 0; i < m_score_arrays.size(); i++) {
        m_score_arrays[i].write(file);
        m_backoff_arrays[i].write(file);
      }
    }
    catch (std::exception &e) {
      throw io_error(std::string("bit::LM::write() failed: ") + e.what());
    }
  }

  void
  LM::read(FILE *file)
  {
    int num_arrays = -1;
    int version = -1;
    int ret = fscanf(file, "LM%d:%d:%d:%d:", &version, 
                     &num_arrays, &m_start_symbol, &m_end_symbol);
    if (ret != 4 || version != 2)
      throw io_error("bit::LM::read() error while reading header");
    m_symbol_map.read(file);
   
    try {
      m_trie.read(file);
      m_score_arrays.resize(num_arrays);
      m_backoff_arrays.resize(num_arrays);
      for (size_t i = 0; i < num_arrays; i++) {
        m_score_arrays[i].read(file);
        m_backoff_arrays[i].read(file);
      }
    }
    catch (std::exception &e) {
      throw io_error(std::string("bit::LM::read() failed: ") + e.what());
    }
  }

  void
  LM::linear_quantization(unsigned int bits)
  {
    if (bits == 0 || bits >= 32)
      throw bit::invalid_argument(
	"bit::LM::linear_quantization() invalid number of bits");
    
    assert(m_score_arrays.size() == m_backoff_arrays.size());
    for (size_t l = 0; l < m_score_arrays.size(); l++) {
      m_score_arrays[l].linear_quantization_bits(bits);
      m_backoff_arrays[l].linear_quantization_bits(bits);
    }
  }

  void 
  LM::separate_leafs(unsigned int level)
  {
    if (m_trie.is_separated(level))
      throw invalid_call("bit::LM::separate_leafs() separated already");
    
    m_trie.separate_leafs(level);
    const CompressedArray &pointer_array = m_trie.pointer_array(level);
    FloatArray &backoff_array = m_backoff_arrays.at(level);
    FloatArray new_backoff_array;
    new_backoff_array.set_mode(backoff_array);

    u32 prev_index = 0;
    for (u64 src = 0; src < pointer_array.num_elems(); src++) {
      u32 index = pointer_array.get(src);
      if (index == prev_index) {
        float backoff = this->backoff(level, src);
        if (backoff != 0) {
          fprintf(stderr, "WARNING: leaf %lld has non-zero backoff %g\n", 
                  src, backoff);
        }
      }
      else {
        assert(index > 0);
        assert(index == prev_index + 1);
        float backoff = this->backoff(level, src);
        if (backoff != 0)
          new_backoff_array.set_grow_widen(index - 1, backoff);
      }
      prev_index = index;
    }
    backoff_array = new_backoff_array;
  }

  void 
  LM::unseparate_leafs(unsigned int level)
  {
    if (!m_trie.is_separated(level))
      throw invalid_call("bit::LM::unseparate_leafs() not separated");
    
    const CompressedArray &pointer_array = m_trie.pointer_array(level);
    FloatArray &backoff_array = m_backoff_arrays.at(level);
    FloatArray new_backoff_array;
    new_backoff_array.set_mode(backoff_array);

    u32 prev_index = 0;
    for (u64 i = 0; i < pointer_array.num_elems(); i++) {
      u32 index = pointer_array.get(i);
      if (index != prev_index) {
        assert(index > 0);
        assert(index == prev_index + 1);
        float backoff = this->backoff(level, index - 1);
        if (backoff != 0)
          new_backoff_array.set_grow_widen(i, backoff);
      }
      prev_index = index;
    }
    backoff_array = new_backoff_array;

    m_trie.unseparate_leafs(level);
  }

  void
  LM::insert_ngram(const std::vector<int> &ngram, float score, float backoff)
  {
    if (ngram.empty())
      throw invalid_argument("lm::LM::insert_ngram(): empty ngram");

    assert(m_start_symbol >= 0);
    assert(m_end_symbol >= 0);

    // Check that insertions are done in sorted order
    //
    if (m_previous_ngram.size() == ngram.size() && 
	compare_ngrams(m_previous_ngram, ngram) >= 0)
    {
      throw invalid_argument(
        std::string("lm::LM::insert_ngram(): sorting error: inserting \"") + 
        ngram_str(ngram) + "\" after \"" + ngram_str(m_previous_ngram) + "\"");
    }
    if (ngram.size() < m_previous_ngram.size() || 
	ngram.size() > m_previous_ngram.size() + 1)    
      throw invalid_argument(
        str::fmt(256, "tried to insert %d-gram \"", (int)ngram.size()) +
	ngram_str(ngram) + 
	str::fmt(256, "\" after %d-gram \"", (int)m_previous_ngram.size()) + 
	ngram_str(m_previous_ngram) + "\"");
    m_previous_ngram = ngram;

    Trie::Iterator it = m_trie.insert_new(ngram);

    // Handle start symbols
    //
    if (ngram.back() == m_start_symbol) {
      if (ngram.size() > 1) {
	fprintf(stderr, "WARNING: skipping ngram: %s\n", 
		ngram_str(ngram).c_str());
	return;
      }
    }

    unsigned int level = ngram.size() - 1;
    if (level >= m_backoff_arrays.size())
      m_backoff_arrays.push_back(FloatArray());
    if (level >= m_score_arrays.size())
      m_score_arrays.push_back(FloatArray());
    u64 index = it.symbol_index();
    m_score_arrays.at(level).set_grow_widen(index, score);
    if (backoff != 0)
      m_backoff_arrays.at(level).set_grow_widen(index, backoff);
  }

  void
  LM::insert_ngram(const std::string &str, float score, float backoff)
  {
    std::vector<std::string> strs(str::split(str::cleaned(str), " \t", true));

    if (strs.empty())
      throw invalid_argument("lm::LM::insert_ngram(): empty ngram");

    std::vector<int> ngram(strs.size());
    for (size_t i = 0; i < strs.size(); i++)
      ngram.at(i) = m_symbol_map.insert(strs.at(i));
    
    insert_ngram(ngram, score, backoff);
  }
  
};
