#ifndef LM_HH
#define LM_HH

#include "bit/Trie.hh"
#include "bit/CompressedArray.hh"
#include "bit/FloatArray.hh"
#include "SymbolMap.hh"

/** A namespace containing classes for storing n-gram language models
 * in a bit-packed trie structure. */
namespace bit {

  /** Class for storing an n-gram language model in a bit-packed trie
   * structure.  Note that the backoff weight must be zero for ngrams
   * that do not have children. 
   *
   * \bug score_arrays waste half of the linear quantization because
   * FloatArray uses sign bit even if all scores are negative.
   */
  class LM {
  public:

    /** Internal Trie type. */
    typedef Trie<CompressedArray> Trie;

    /** Iterator type. */
    typedef Trie::Iterator Iterator;

    /** Type for symbol map. */
    typedef SymbolMap<std::string, int> SymbolMap;

    /** Default constructor. */
    LM()
    {
      reset();
    }

    /** Reset the model to initial state. */
    void reset()
    {
      m_symbol_map = SymbolMap();
      m_start_symbol = -1;
      m_end_symbol = -1;
      m_trie = Trie();
      m_backoff_arrays.clear();
      m_score_arrays.clear();
      m_previous_ngram.clear();
    }

    /** Order of the model. */
    unsigned int order() const
    {
      return m_score_arrays.size();
    }

    /** The number of bytes required to store all bit-buffers. */
    u64 size() const
    {
      u64 size = 0;
      assert(m_score_arrays.size() == m_backoff_arrays.size());
      for (size_t i = 0; i < m_score_arrays.size(); i++) {
        size += m_score_arrays[i].data_len();
        size += m_backoff_arrays[i].data_len();
      }
      return size + m_trie.size();
    }

    /** Access to score arrays. */
    const FloatArray &score_array(unsigned int level) const
    {
      return m_score_arrays.at(level);
    }

    /** Access to backoff arrays. */
    const FloatArray &backoff_array(unsigned int level) const
    {
      return m_backoff_arrays.at(level);
    }

    /** Access to symbol arrays of the trie. */
    const CompressedArray &symbol_array(unsigned int level) const
    {
      return m_trie.symbol_array(level);
    }

    /** Access to pointer arrays of the trie. */
    const CompressedArray &pointer_array(unsigned int level) const
    {
      return m_trie.pointer_array(level);
    }

    /** Access to child limit arrays of the trie. */
    const CompressedArray &child_limit_array(unsigned int level) const
    {
      return m_trie.child_limit_array(level);
    }

    /** Read language model from file in ARPA format. 
     * \param file = file stream to read from
     * \param sentence_start_str = the label of sentence start symbol
     * \param sentence_end_str = the label of sentence end symbol
     */
    void
    read_arpa(FILE *file, const std::string &sentence_start_str = "<s>",
              const std::string &sentence_end_str = "</s>", 
              bool verbose = false);

    /** Write the model in ARPA format. 
     * \param FILE = file stream to write to
     */
    void write_arpa(FILE *file) const;

    /** Write the model in binary format. 
     * \param FILE = file stream to write to
     * \throw bit::io_error on error
     */
    void write(FILE *file) const;

    /** Read the model from file stored in binary format. 
     * \param FILE = file stream to write to
     * \throw bit::io_error on error
     */
    void read(FILE *file);

    /** Quantize all floats linearly.  Does nothing on levels that are
     * already quantized.
     *
     * \param bits = number of bits per float (remember that sign
     * requires a bit)
     *
     * \throw bit::invalid_argument if bits < 2 or bits > 32.
     */
    void linear_quantization(unsigned int bits);

    /** Compress the arrays of the trie on the given level. 
     * \param level = the level to compress
     */
    void compress_trie(unsigned int level)
    {
      m_trie.compress(level);
    }

    /** Compress all levels of the trie. */
    void compress_trie()
    {
      m_trie.compress();
    }

    /** Uncompress one level of the trie. 
     * \param level = the level to uncompress
     */
    void uncompress_trie(unsigned int level)
    {
      m_trie.uncompress(level);
    }

    /** Uncompress the trie. */
    void uncompress_trie()
    {
      m_trie.uncompress();
    }

    /** Separate leafs of the trie on a given level, and modify the
     * backoff array accordingly. 
     *
     * \throw bit::invalid_call if level is separated already
     *
     * \throw bit::out_of_range if \c level is greater or equal to highest
     * level
     */
    void separate_leafs(unsigned int level);

    /** Unseparate leafs (and remove possible compression) of the trie
     * on a given level, and modify the backoff array accordingly.
     *
     * \throw bit::invalid_call if level is not separated
     *
     * \throw bit::out_of_range if \c level is greater or equal to highest
     * level
     */
    void unseparate_leafs(unsigned int level);

    /** Insert a new ngram to the model.
     * \param ngram = vector of ngram symbols
     * \param score = score (usually log-probability) of the ngram
     * \param backoff = the backoff weight of the ngram
     *
     * \throw bit::invalid_argument if ngrams are not inserted in
     * sorted order or if ngram is empty
     */
    void
    insert_ngram(const std::vector<int> &ngram, float score, float backoff);

    /** Insert a new ngram to the model allowing to insert new symbols.
     * \param str = white-space separated list of symbols
     * \param score = score (usually log-probability) of the ngram
     * \param backoff = the backoff weight of the ngram
     *
     * \throw bit::invalid_argument if ngrams are not inserted in
     * sorted order or if ngram is empty
     */
    void
    insert_ngram(const std::string &str, float score, float backoff);

    /** Set sentence start symbol and add the string in symbol mapping
     * if not there already.
     * \throw bit::invalid_call if set already
     */
    void set_start_symbol(const std::string &str)
    {
      if (m_start_symbol >= 0)
        throw bit::invalid_call("bit::LM::set_start_symbol() called again");
      m_start_symbol = m_symbol_map.insert(str);
    }

    /** Set sentence end symbol and add the string in symbol mapping
     * if not there already.
     * \throw bit::invalid_call if set already
     */
    void set_end_symbol(const std::string &str)
    {
      if (m_end_symbol >= 0)
        throw bit::invalid_call("bit::LM::set_end_symbol() called again");
      m_end_symbol = m_symbol_map.insert(str);
    }

    /** The symbol starting the sentence. */
    int start_symbol() const { 
      return m_start_symbol;
    }

    /** The symbol ending the sentence. */
    int end_symbol() const { 
      return m_end_symbol;
    }

    /** The mapping between symbols and strings. */
    const SymbolMap &symbol_map() const
    {
      return m_symbol_map;
    }

    /** Printable string of a ngram. 
     * \param vec = ngram to convert
     * \return string representing the ngram
     */
    template <class T>
    std::string 
    ngram_str(const std::vector<T> &vec) const
    {
      assert(!vec.empty());
      std::string str;
      for (size_t o = 0; o < vec.size(); o++) {
        if (o != 0)
          str.append(" ");
        str.append(m_symbol_map.at(vec[o]));
      }
      return str;
    }

    /** Trie iterator pointing to the root. */
    Iterator root() const
    {
      return Iterator(m_trie);
    }

    /** Backoff weight at the iterator position. 
     * \throw bit::invalid_call if called at root
     */
    float backoff(const Iterator &it) const
    {
      if (it.is_root())
        throw invalid_call("lm::LM::backoff() called at root");
      u32 index = it.child_limit_index();
      if (index == max_u32)
        return 0;
      return backoff(it.length() - 1, index);
    }

    /** Backoff weight from a given level. 
     * \param level = level to access
     * \param index = index on the level
     * \return the backoff value
     * \throw bit::invalid_argument if level exceeds levels
     * \throw bit::invalid_argument if index exceeds number of elements
     */
    float backoff(unsigned int level, u64 index) const
    {
      if (level >= m_score_arrays.size())
        throw bit::invalid_argument("bit::LM::backoff() level too high");
      const FloatArray &score_array = m_score_arrays.at(level);
      const FloatArray &backoff_array = m_backoff_arrays.at(level);
      if (index >= score_array.num_elems())
        throw bit::invalid_argument("bit::LM::backoff() index too high");
      if (index >= backoff_array.num_elems())
        return 0;
      return backoff_array.get(index);
    }

    /** Probability score at the iterator position. */
    float score(const Iterator &it) const
    {
      u64 index = it.symbol_index();
      unsigned int level = it.length() - 1;
      return m_score_arrays.at(level).get(index);
    }

    /** Walk iterator to given symbol backoffing if necessary.  It is
     * also ensured that the resulting iterator has children,
     * backoffing if necessary.
     * 
     * \param it = iterator to walk
     * \param symbol = symbol to find
     * \return the score of the walked path
     */
    float walk(Iterator &it, int symbol) const
    {
      float score = 0;
      while (!it.goto_child(symbol)) {
        assert(it.length() > 0);
        score += backoff(it);
        it.goto_backoff_full();
      }
      u64 index = it.symbol_index();
      score += m_score_arrays.at(it.length() - 1).get(index);
      if (it.num_children() == 0) {
	assert(backoff(it) == 0);
        it.goto_backoff_full();
      }
      return score;
    }

  private:

    /** Compare two ngrams.  
     * \return -1 if \c a comes before \c b, 0 if equal, 1 if \c a
     * comes after \c b */
    int compare_ngrams(const std::vector<int> &a, const std::vector<int> &b)
    {
      std::vector<int>::size_type o = 0;
      while (1) {
        if (o == a.size() && o == b.size())
          return 0;
        if (o == a.size())
          return -1;
        if (o == b.size())
          return 1;
        if (a[o] < b[o])
          return -1;
        if (a[o] > b[o])
          return 1;
        o++;
      }
    }

  private:

    /** Mapping between model symbols strings and symbols. */
    SymbolMap m_symbol_map;

    /** Symbol corresponding to the sentence start symbol. */
    int m_start_symbol;

    /** Symbol corresponding to the sentence end symbol. */
    int m_end_symbol;

    /** The internal Trie structure. */
    Trie m_trie;

    /** Arrays containing backoff weights for each n-gram level. */
    std::vector<FloatArray> m_backoff_arrays;

    /** Arrays containing probability scores for each n-gram level. */
    std::vector<FloatArray> m_score_arrays;

    /** Previous ngram inserted in the model. */
    std::vector<int> m_previous_ngram;
  };

};

#endif /* LM_HH */
