#ifndef PERPLEXITY_HH
#define PERPLEXITY_HH

#include "lm/LM.hh"

namespace bit {
  
  /** Compute perplexity or cross-entropy of a language model. */
  class Perplexity {
  public:
    
    /** Create a new perplexity computer with a language model. */
    Perplexity(const LM &lm)
      : m_lm(&lm)
    {
      opt.word_boundary_str = "<w>";
      opt.unk_str = "";
      reset();
    }

    /** Reset perplexity counters. */
    void reset()
    {
      m_start_pending = true;
      m_score = 0;
      m_num_symbols = 0;
      m_num_words = 0;
      m_num_sentences = 0;
      m_it = m_lm->root();
    }

    /** Number of symbols. */
    int num_symbols() const
    {
      return m_num_symbols;
    }

    /** Number of words. */
    int num_words() const
    {
      return m_num_words;
    }

    /** Log-probability of the test data. */
    double score() const
    {
      return m_score;
    }

    /** Compute cross-entropy in bits assuming that score is in
     * log10. 
     * \throw bit::invalid_call if no words yet
     */
    float cross_entropy_per_word() const
    {
      if (m_num_words == 0)
        throw bit::invalid_call(
          "bit::Perplexity::cross_entropy_per_word() no words yet");
      return m_score * 3.3219280949 / m_num_words;
    }

    /** Add a symbol to eval set.
     *
     * \param symbol = symbol to add
     * \return log10-probability of the word
     *
     * \throw bit::invalid_argument if sentence start was missing or
     * sentence start was given before sentence end, or unknown symbol
     * was given
     */
    float add_symbol(const std::string &symbol_str)
    {
      int symbol = m_lm->symbol_map().index_nothrow(symbol_str);
      bool unk = false;
      if (symbol < 0) {
        if (opt.unk_str.empty()) 
          throw bit::invalid_argument(
            std::string("bit::Perplexity::add_symbol(): invalid symbol \"") +
            symbol_str + "\"");
        symbol = m_lm->symbol_map().index_nothrow(opt.unk_str);
        if (symbol < 0) {
          throw bit::invalid_argument(
            std::string("bit::Perplexity::add_symbol(): invalid unk symbol \"")
            + opt.unk_str + "\"");
        }
        unk = true;
      }
      
      if (symbol == m_lm->start_symbol() && !m_start_pending)
        throw bit::invalid_argument(
          "bit::Perplexity::add_symbol() unexpected start symbol");
      if (symbol != m_lm->start_symbol() && m_start_pending)
        throw bit::invalid_argument(
          "bit::Perplexity::add_symbol() expected start symbol but got \"" +
          symbol_str + "\"");

      float score = m_lm->walk(m_it, symbol);

      if (symbol == m_lm->start_symbol()) {
        m_start_pending = false;
        assert(score == 0);
      }
      else if (symbol != m_lm->end_symbol() && !unk) {
        m_num_symbols++;
      }

      if (!unk)
        m_score += score;

      if (symbol_str == opt.word_boundary_str)
        m_num_words++;

      if (symbol == m_lm->end_symbol()) {
        m_start_pending = true;
        m_num_sentences++;
        m_num_words--;
      }

      return score;
    }

  public:
    struct {
      /** Symbol used for computing number of words.  It is assumed
       * that word boundary comes always after sentence start and
       * before sentence end. */
      std::string word_boundary_str;

      /** Symbol used for unknown word.  Empty if unk is not used. */
      std::string unk_str;
    } opt;

  private:

    /** Are we waiting the start of sentence? */
    bool m_start_pending;

    /** Score of the test data. */
    double m_score;

    /** Number of symbols in the test data (not including sentence starts) */
    int m_num_symbols;

    /** Number of words in the test data. */
    int m_num_words;

    /** Number of sentences in the test data. */
    int m_num_sentences;

    /** The language model used for computing the perplexity. */
    const LM *m_lm;

    /** Iterator for walking in the language model. */
    LM::Iterator m_it;

  };

};

#endif /* PERPLEXITY_HH */
