#ifndef STR_HH
#define STR_HH

#include <string>
#include <vector>
#include <stdexcept>

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <math.h>

/** Functions for handling strings of the Standard Template Library.
 * \bug We assume that the newline is '\n', which can cause problems
 * in Windows.
 */
namespace str {

  /** Format a string. 
   * \param size = maximum size of the resulting string (cut rest)
   * \param fmt = the format string as in standard printf()
   * \return the formatted string
   */
  inline std::string fmt(size_t size, const char *fmt, ...)
  {
    va_list ap;
    va_start(ap, fmt);
    char *buf = new char[size];
    vsnprintf(buf, size, fmt, ap);
    va_end(ap);

    std::string str(buf);
    delete[] buf;
    return str;
  }

  /** Remove leading and trailing characters from a string. 
   * \param str = the string to be cleaned
   * \param chars = the characters to clean from the string
   */
  inline void clean(std::string &str, const char *chars = " \n\t")
  {
    int i;

    // Clean trailing chars
    //
    for (i = str.length(); i > 0; i--)
      if (!strchr(chars, str[i - 1]))
	break;
    str.erase(i);

    // Clean leading chars
    //
    for (i = 0; i < (int)str.length(); i++)
      if (!strchr(chars, str[i]))
	break;
    str.erase(0, i);
  }

  /** Remove leading and trailing characters from a string. 
   * \param str = the string to be cleaned
   * \param chars = the characters to clean from the string
   * \return the cleaned string
   */
  inline std::string cleaned(const std::string &str, 
                             const char *chars = " \n\t")
  {
    std::string ret = str;
    clean(ret, chars);
    return ret;
  }

  /** Split a string to fields.  If \c num_fields is positive, it
   * specifies the maximum number of fields.  If the line contains
   * more fields, they are included in the last field as such (also
   * possible delimiters without processing).
   *
   * \param str = the string to be splitted
   * \param delims = the delimiter characters
   * \param group = group subsequent delimiters as one delimeter
   * \param num_fields = the maximum number of fields to return
   * \return fields the vector containing the resulting fields
   */
  inline std::vector<std::string>
  split(const std::string &str, const char *delims, bool group, 
        unsigned int num_fields = 0)
  {
    std::vector<std::string> fields;
    size_t begin = 0;
    size_t end = 0;

    fields.clear();
    bool delim_pending = false;
    while (begin < str.length()) {
      // If 'fields' fields was requested and, this is the last field,
      // include the rest.
      if (num_fields > 0 && fields.size() == num_fields - 1) {
        delim_pending = false;
	fields.push_back(str.substr(begin));
	break;
      }

      // Find the string before the next delim
      delim_pending = false;
      while (end < str.length()) {
        if (strchr(delims, str[end])) {
          delim_pending = true;
          break;
        }
	end++;
      }
      fields.push_back(str.substr(begin, end - begin));

      // Eat the delim or group of delims
      end++;
      if (group)
	while (end < str.length() && strchr(delims, str[end]))
	  end++;

      begin = end;
    }

    if (delim_pending)
      fields.push_back("");

    return fields;
  }

  /** Convert a string to signed long.  Arbitrary amount of white
   * space is accepted before the number as explained on strtol(3)
   * manual page.  Anything after the number is considered as an
   * conversion error.
   *
   * \param str = the string to convert
   * \param base = base of the coversion
   * \throw std::out_of_range if argument was out of range
   * \throw std::invalid_argument if conversion error occured
   */
  inline long str2long(const std::string &str, int base = 10)
  {
    char *endptr;
    const char *c_str = str.c_str();

    long value = strtol(c_str, &endptr, base);
    if (value == LONG_MIN || value == LONG_MAX)
      throw std::out_of_range("str::str2long: argument out of range");
    if (*c_str == '\0' || *endptr != '\0')
      throw std::invalid_argument("str::str2long: invalid string");

    return value;
  }

  /** Convert a string to unsigned long.  Arbitrary amount of white
   * space is accepted before the number as explained on strtol(3)
   * manual page.  Anything after the number is considered as an
   * conversion error.
   *
   * \param str = the string to convert
   * \param base = base of the coversion
   * \throw std::out_of_range if argument was out of range
   * \throw std::invalid_argument if conversion error occured
   */
  inline unsigned long str2ulong(const std::string &str, int base = 10)
  {
    char *endptr;
    const char *c_str = str.c_str();

    unsigned long value = strtoul(c_str, &endptr, base);
    if (value == ULONG_MAX)
      throw std::out_of_range("str::str2ulong: argument out of range");
    if (*c_str == '\0' || *endptr != '\0')
      throw std::invalid_argument("str::str2ulong: invalid string");

    return value;
  }

  /** Convert a string to float.  Arbitrary amount of white space is
   * accepted before the float as explained on strtof(3) manual page.
   * Anything after the number is considered as an conversion error.
   *
   * \param str = the string to convert
   * \throw std::out_of_range if argument was out of range
   * \throw std::invalid_argument if conversion error occured
   */
  inline float str2float(const std::string &str)
  {
    char *endptr;
    const char *c_str = str.c_str();

    float value = strtof(c_str, &endptr);
    if (value == -HUGE_VALF || value == HUGE_VALF)
      throw std::out_of_range("str::str2float: argument out of range");
    if (*c_str == '\0' || *endptr != '\0')
      throw std::invalid_argument("str::str2float: invalid string");

    return value;
  }


  /** Convert a string to a vector of integers using str2long(). 
   * \param str = the string to convert
   * \return a vector of integers
   * \throw exceptions as described in str2long()
   */
  template <class T>
  std::vector<T> long_vec(const std::string &str)
  {
    std::vector<std::string> str_vec(
      str::split(str::cleaned(str), " \t\n", true));
    std::vector<T> vec(str_vec.size());
    for (size_t i = 0; i < str_vec.size(); i++)
      vec[i] = str::str2long(str_vec[i]);
    return vec;
  }

  /** Convert a string to a vector of floats using str2float(). 
   * \param str = the string to convert
   * \return a vector of floats
   * \throw exceptions as described in str2float()
   */
  inline std::vector<float> float_vec(const std::string &str)
  {
    std::vector<std::string> str_vec(
      str::split(str::cleaned(str), " \t\n", true));
    std::vector<float> vec(str_vec.size());
    for (size_t i = 0; i < str_vec.size(); i++)
      vec[i] = str::str2float(str_vec[i]);
    return vec;
  }

  /** Return a strin with the possible trailing newline removed
   * \param str = string to homp
   * \param chars = characters to remove
   * \return the string without trailing newline
   */
  inline std::string chomped(const std::string &str, const char *chars = "\n")
  {
    if (!str.empty() && strchr(chars, str[str.length() - 1]))
      return str.substr(0, str.length() - 1);
    return str;
  }

  /** Remove the possible trailing newline from a string
   * \param str = string to homp
   * \param chars = characters to remove
   * \return reference to modified \c str 
   */
  inline std::string &chomp(std::string &str, const char *chars = "\n")
  {
    if (!str.empty() && strchr(chars, str[str.length() - 1]))
      str.erase(str.end() - 1);
    return str;
  }

  /** Read a line from a file.
   * \param str = the string to read to
   * \param file = the file to read from
   * \param do_chomp = should we chomp the possible newline at the end
   * \return false if end of file reached (\c str will be empty)
   * \throw str::io_error if read failed
   */
  inline bool 
  read_line(std::string &str, FILE *file = stdin, bool do_chomp = false)
  {
    char *ptr;
    char buf[4096];

    str.erase();
    while (1) {
      ptr = fgets(buf, 4096, file);
      if (!ptr)
	break;

      str.append(buf);
      if (str[str.length() - 1] == '\n') {
	break;
      }
    }

    if (ferror(file) || str.length() == 0)
      return false;

    if (do_chomp)
      chomp(str, "\n");

    return true;
  }


};

#endif /* STR_HH */
