#include "conf/conf.hh"
#include "io/io.hh"
#include "str/str.hh"
#include "kn/Kneser.hh"

using namespace bit;

conf::Config config;

int
main(int argc, char *argv[])
{
  try {
    config("usage: knprune [OPTION...]\n")
      ('h', "help", "", "", "display help")
      ('\0', "counts=FILE", "arg must", "-", "read counts from file")
      ('\0', "discounts=STR", "arg must", "", 
       "white-space separated list of discounts")
//      ('\0', "prune-limit=INT", "arg must", "", 
//       "number of ngrams to leave in the model")
      ('\0', "prune-threshold=FLOAT", "arg", "", "pruning threshold")
      ('\0', "prune-list=FILE", "arg", "", 
       "file containing ngrams to be pruned (one per line)")

      ('\0', "reserve-orders=INT", "arg", "10", 
       "reserve orders to avoid growing arrays (default 10)")
      ('\0', "out-arpa=FILE", "arg", "-", "output ARPA language model")

      ('\0', "read-counts=FILE", "arg", "", 
       "read count structures (binary) ")
      ('\0', "write-counts=FILE", "arg", "", 
       "write count structures (binary) ")
      ('\0', "read-d1d2=FILE", "arg", "", 
       "read d1 and d2 structures (binary) ")
      ('\0', "write-d1d2=FILE", "arg", "", 
       "write d1 and d2 structures (binary) ")
      ('\0', "d1-model=STR", "arg", "kn", 
       "model used in d1 computation: kn, abs (default kn)")
      ('\0', "d1-weight-model=STR", "arg", "kn", 
       "weight model used in d1 computation: kn, abs (default kn)")

      ('\0', "debug-write-counts=FILE", "arg", "", 
       "write counts in ascii")
      ;
    config.default_parse(argc, argv);
    if (config.arguments.size() != 0)
      config.print_help(stderr, 1);

    if (config["prune-list"].specified && config["prune-threshold"].specified)
      throw std::string("both --prune-threshold and --prune-list given");

    bool read_binary_counts = config["read-counts"].specified;
    bool read_binary_d1d2 = config["read-d1d2"].specified;

    // Read the counts
    //
    Kneser k;

    std::string d1_model = config["d1-model"].get_str();
    std::string d1_weight_model = config["d1-weight-model"].get_str();
    if (d1_model == "kn")
      k.set_d1_model(0);
    else if (d1_model == "abs")
      k.set_d1_model(1);
    else {
      fprintf(stderr, "invalid d1-model\n");
      exit(1);
    }

    if (d1_weight_model == "kn")
      k.set_d1_weight_model(0);
    else if (d1_weight_model == "abs")
      k.set_d1_weight_model(1);
    else {
      fprintf(stderr, "invalid d1-weight-model\n");
      exit(1);
    }

    k.reserve_orders(config["reserve-orders"].get_int());

    if (read_binary_counts) {
      fprintf(stderr, "reading binary counts from %s\n", 
              config["read-counts"].get_c_str());
      k.read_binary_counts(io::Stream(config["read-counts"].get_str()));
    }
    else {
      fprintf(stderr, "reading counts from %s\n", 
              config["counts"].get_c_str());
      k.read_counts(io::Stream(config["counts"].get_str()), false);
    }    

    fprintf(stderr, "got %lld ngrams\n", k.num_ngrams());

    if (config["discounts"].specified) {
      std::vector<float> vec(str::float_vec(config["discounts"].get_str()));
      fprintf(stderr, "using discounts:");
      for (size_t i = 0; i < vec.size(); i++) {
        fprintf(stderr, " %g", vec[i]);
        k.set_discount(i, vec[i]);
      }
      fputs("\n", stderr);
    }

    if (read_binary_counts)
      fprintf(stderr, "using modified counts from the binary file\n");
    else
      k.compute_sums();

    if (config["write-counts"].specified) {
      fprintf(stderr, "writing counts in binary file: %s\n",
              config["write-counts"].get_c_str());
      k.write_binary_counts(io::Stream(config["write-counts"].get_str(), "w"));
    }

    if (read_binary_d1d2) {
      fprintf(stderr, "reading binary d1 and d2 from %s\n", 
              config["read-d1d2"].get_c_str());
      k.read_binary_d1d2(io::Stream(config["read-d1d2"].get_str()));
    }
    else {
      k.compute_d1();
      k.compute_d2_full();
    }

    if (config["write-d1d2"].specified) {
      fprintf(stderr, "writing d1 and d2 in binary file: %s\n",
              config["write-d1d2"].get_c_str());
      k.write_binary_d1d2(io::Stream(config["write-d1d2"].get_str(), "w"));
    }

    if (config["prune-threshold"].specified) {
      float prune_threshold = config["prune-threshold"].get_float();
      fprintf(stderr, "pruning with threshold %g\n", prune_threshold);
      k.prune_threshold(prune_threshold);
    }
    else if (config["prune-list"].specified) {
      fprintf(stderr, "pruning ngrams from %s\n", 
              config["prune-list"].get_c_str());
      io::Stream prune_stream(config["prune-list"].get_str());
      std::string line;
      Progress p(7951);
      p.set_report_string("pruning:");
      while (str::read_line(line, prune_stream, true)) {
        p.step();
        str::clean(line);
        if (line.empty())
          continue;
        Kneser::Iterator it = k.find(line);
        if (!k.is_pruned(it))
          k.prune_ngram(k.find(line));
      }
      p.finish();
    }
    
    k.compute_beta_numerator_terms();
    k.compute_beta_denominator();
    k.compute_beta_interpolation_numerator();
    fprintf(stderr, "pruning finished\n");

    fprintf(stderr, "writing %s\n", config["out-arpa"].get_c_str());
    k.write_beta_arpa(io::Stream(config["out-arpa"].get_str(), "w"));

    if (config["debug-write-counts"].specified) {
      fprintf(stderr, "writing debug counts in %s\n", 
              config["debug-write-counts"].get_c_str());
      k.debug_write_counts(
        io::Stream(config["debug-write-counts"].get_str(), "w"));
    }
  }
  catch (std::string &str) {
    fprintf(stderr, "exception: %s\n", str.c_str());
    exit(1);
  }
  catch (std::exception &e) {
    fprintf(stderr, "exception: %s\n", e.what());
    exit(1);
  }
}

