#include "conf/conf.hh"
#include "io/io.hh"
#include "str/str.hh"
#include "lm/Perplexity.hh"
#include "Timer.hh"

using namespace bit;

conf::Config config;
LM lm;

int
main(int argc, char *argv[])
{
  try {
    config("usage: preplexity [OPTION...]\n")
      ('h', "help", "", "", "display help")
      ('\0', "arpa=FILE", "arg", "", "read ARPA language model")
      ('\0', "blm=FILE", "arg", "", "read binary language model")
      ('\0', "text=FILE", "arg", "-", "evaluation text")
      ('\0', "num-repeats=INT", "arg", "1", 
       "number of times to repeat computation (for robust time measurements)")
      ('\0', "wait-key", "", "", "wait keypress before exiting")
      ('\0', "quiet", "", "", "do not print sentence scores")
      ('\0', "use-unk=STR", "arg", "", "use unknown symbol")
      ;
    config.default_parse(argc, argv);
    if (config.arguments.size() != 0)
      config.print_help(stderr, 1);

    // Read the language model
    //
    fprintf(stderr, "reading the language model\n");
    if (config["arpa"].specified) {
      if (config["blm"].specified) {
        fprintf(stderr, "options --arpa and --blm not allowed together\n");
        exit(1);
      }
      lm.read_arpa(io::Stream(config["arpa"].get_str()), "<s>", "</s>", true);
    }
    else if (config["blm"].specified) {
      lm.read(io::Stream(config["blm"].get_str()));
    }
    else {
      fprintf(stderr, "option --arpa or --blm required\n");
      exit(1);
    }
    fprintf(stderr, "model order %d\n", lm.order());
    fprintf(stderr, "model size %lld\n", lm.size());

    io::Stream in(config["text"].get_str());
    std::string line;
    std::vector<std::string> words;

    typedef std::vector<std::string> Sentence;
    std::vector<Sentence> text;

    // Read all test words
    //
    fprintf(stderr, "reading the evaluation text\n");
    while (str::read_line(line, in, true)) {
      words = str::split(line, " \t", true);
      if (words.empty())
        continue;
      if (words[0] == "<s>")
        words.erase(words.begin());
      if (words.back() != "</s>")
	words.push_back("</s>");

      text.push_back(words);
    }

    std::string unk_str = config["use-unk"].get_str();
     
    // Compute perplexity
    //
    fprintf(stderr, "computing perplexity\n");
    Timer timer;
    timer.start();
    int num_rounds = config["num-repeats"].get_int();
    bool quiet = config["quiet"].specified;
    for (int round = 0; round < num_rounds; round++) {

      double total_score = 0;
      int num_symbols = 0;
      int num_words = 0;

      for (size_t s = 0; s < text.size(); s++) {
        Perplexity p(lm);
        p.opt.unk_str = unk_str;
        Sentence &sent = text[s];
        
        p.add_symbol("<s>");
        for (size_t i = 0; i < sent.size(); i++)
          p.add_symbol(sent[i]);
        total_score += p.score();
        if (!quiet) {
          printf("sentence score: %g\n", p.score());
          fflush(stdout);
        }
        num_symbols += p.num_symbols();
        num_words += p.num_words();
      }

      printf("total score: %g\n", total_score);
      printf("symbols: %d\n", num_symbols);
      printf("symbol perplexity: %g\n", 
             exp(-total_score * log(10) / num_symbols));
      printf("word perplexity (for morphs only): %g\n", 
             exp(-total_score * log(10) / num_words));
      printf("cross entropy (bits per word boundaries): %g\n", 
             -total_score * 3.3219280949 / num_words);
      fflush(stdout);
    }
    timer.stop();

    fprintf(stderr, "elapsed CPU time: %g sec\n", timer.total_sec());

    if (config["wait-key"].specified)
      fgetc(stdin);
  }
  catch (std::string &str) {
    fprintf(stderr, "exception: %s\n", str.c_str());
    exit(1);
  }
  catch (std::exception &e) {
    fprintf(stderr, "exception: %s\n", e.what());
    exit(1);
  }
}

